# -*- coding: utf-8 -*-
# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import json
import os
import pytest
import tempfile

from mock import call, patch
from tests import TEST_DATA_DIR, MockMBSBuildsData, make_mmd
from ursa_major.cli import main


class TestCheckConfig:

    def setup_method(self, method):
        fd, self.tag_config_file = tempfile.mkstemp()
        os.close(fd)

        self.ClientSession_patcher = patch('koji.ClientSession')
        self.mock_ClientSession = self.ClientSession_patcher.start()
        self.koji_session = self.mock_ClientSession.return_value

        self.get_patcher = patch('requests.get')
        self.mock_get = self.get_patcher.start()

    def teardown_method(self, method):
        self.get_patcher.stop()
        os.unlink(self.tag_config_file)

    def write_tag_config_file(self, content):
        with open(self.tag_config_file, 'w') as f:
            json.dump(content, f)

    def check_config(self):
        config_file = os.path.join(TEST_DATA_DIR, 'ursa-major-test.conf')
        cli_cmd = [
            'ursa-major', 'check-config',
            '--config', config_file,
            '--tag-config-file', self.tag_config_file
        ]

        with patch('sys.argv', new=cli_cmd):
            main()

    @pytest.mark.parametrize('modules,error_log', [
        ([
            [{'name': 'mariadb'}],
            'Missing stream in module config: {}'.format({u'name': u'mariadb'})
        ]),
        ([
            [{'stream': '10.4', 'priority': 10}],
            'Missing name in module config: {}'.format(
                {u'stream': u'10.4', u'priority': 10})
        ]),
        ([
            [{'name': 'mariadb', 'stream': '10.4'}],
            'Missing priority in module config: {}'.format(
                {u'name': u'mariadb', u'stream': u'10.4'})
        ]),
    ])
    def test_missing_mandatory_fields(self, modules, error_log):
        self.write_tag_config_file({
            'f30-build': {
                'modules': modules,
                'owners': ['owner@example.com']
            }
        })

        self.koji_session.getInheritanceData.return_value = [
            {
                'parent': 10,
                'name': 'module-ant-1.10-20181122140939-819b5873',
                'priority': 10
            }
        ]

        with patch('ursa_major.handlers.check_config.log') as log:
            # This is only for making the program terminate and raise SystemExit
            log.error.counter = 3

            with pytest.raises(SystemExit):
                self.check_config()

            log.error.assert_has_calls([call(error_log)])

    def test_show_error_if_cannot_get_tag_inheritance_data(self):
        self.write_tag_config_file({
            'f30-build': {
                'modules': [],
                'owners': ['owner@example.com']
            }
        })

        # This test requires an error raised from Koji API getInheritanceData
        # So, just raise any error
        self.koji_session.getInheritanceData.side_effect = ValueError

        with patch('ursa_major.handlers.check_config.log') as log:
            # This is only for making the program terminate and raise SystemExit
            log.error.counter = 3

            with pytest.raises(SystemExit):
                self.check_config()

            log.error.assert_has_calls([
                call('Unable to get inheritance data of tag %s: %s', u'f30-build', '')
            ], any_order=True)

    def test_verify_valid_module_config(self):
        self.write_tag_config_file({
            'f30-build': {
                'modules': [
                    {
                        'name': 'mariadb', 'stream': '10.4', 'priority': 30,
                        'requires': {'platform': 'f30'},
                        'buildrequires': {'platform': 'f30'}
                    },
                    {'name': 'ffsend', 'stream': 'latest', 'priority': 40},
                ],
                'owners': ['owner@example.com']
            }
        })

        mock_builds = [
            {
                'id': 3617,
                'name': 'mariadb',
                'koji_tag': 'module-mariadb-10.4-3020190313091759-a5b0195c',
                'modulemd': make_mmd(name='mariadb', stream='10.4',
                                     version='3020190313091759', context='a5b0195c',
                                     requires={'platform': 'f30'},
                                     buildrequires={'platform': 'f30'}).dumps(),
            },
            {
                'id': 3640,
                'name': 'ffsend',
                'koji_tag': 'module-ffsend-latest-3020190316024948-a5b0195c',
                'modulemd': make_mmd(name='mariadb', stream='10.4',
                                     version='3020190316024948', context='a5b0195c',
                                     requires={'platform': 'f30'},
                                     buildrequires={'platform': 'f30'}).dumps(),
            }
        ]

        mock_mbs_builds_data = MockMBSBuildsData(mock_builds)
        self.mock_get.side_effect = mock_mbs_builds_data.get

        self.koji_session.getInheritanceData.return_value = [
            {
                'parent': 10,
                'name': 'module-ant-1.10-20181122140939-819b5873',
                'priority': 10
            }
        ]

        with patch('ursa_major.handlers.check_config.log') as log:
            # This is only for making the program terminate normally
            log.error.counter = 0

            self.check_config()
            log.error.assert_not_called()



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import json
import mock
import os
import shutil
import tempfile
from argparse import Namespace

from tests import UrsaMajorTestCase, MockResponse, make_mmd
from ursa_major import ModuleConfig, MBS_BUILD_STATES
from ursa_major.handlers.add_tag import ModuleInfo, AddTagHandler
from ursa_major.mbs import MBS


class AddTagHandlerTestCase(UrsaMajorTestCase):
    def setUp(self):

        config = mock.MagicMock()
        self.handler = AddTagHandler(config)

        self.tmpdir = tempfile.mkdtemp(suffix='_ursa_major_test')
        self.tag_config_file = os.path.join(self.tmpdir, 'default.json')
        with open(self.tag_config_file, 'w') as f:
            json.dump({}, f)

        self.set_args()

    def tearDown(self):
        try:
            shutil.rmtree(self.tmpdir)
        except:  # noqa
            pass

    def write_tag_config(self, content):
        """ Write content to tag config file.

        :param content: content in dict format
        """
        with open(self.tag_config_file, 'w') as f:
            json.dump(content, f, indent=2, sort_keys=True)

    def set_args(self, **kwargs):
        kwargs.setdefault('tag_config_file', self.tag_config_file)
        kwargs.setdefault('module_from_env', 'CI_MESSAGE')
        kwargs.setdefault('wait_regen_repo', False)
        kwargs.setdefault('send_mail', False)
        kwargs.setdefault('dry_run', False)
        args = Namespace()
        for k, v in kwargs.items():
            setattr(args, k, v)
        self.handler.set_args(args)


class TestAddTagHandler(AddTagHandlerTestCase):
    def setUp(self):
        super(TestAddTagHandler, self).setUp()
        fake_tag_config = {
            'rhel-8.0-build': {
                'owners': ['foo@example.com'],
                'modules': [
                    {'name': 'autotools', 'stream': 'rhel-8.0', 'priority': 10,
                     'requires': {'platform': 'rhel-8.0'}},
                    {'name': 'ant', 'stream': '1.10', 'priority': 20,
                     'requires': {'platform': 'rhel-8.0'}},
                ]
            },
            'other-8.0-tag': {
                'owners': ['bar@example.com'],
                'modules': [
                    {'name': 'autotools', 'stream': 'rhel-8.0', 'priority': 30,
                     'requires': {'platform': 'rhel-8.0'}}
                ]
            }
        }
        self.write_tag_config(fake_tag_config)

    @mock.patch('ursa_major.handlers.add_tag.log')
    @mock.patch('ursa_major.handlers.base.KojiService')
    def test_terminate_if_module_build_is_not_ready(self, KojiService, log):
        ci_message = json.dumps(dict(
            id=1,
            state=4,
            state_name='done',
        ))

        with mock.patch.dict(os.environ, {'CI_MESSAGE': ci_message}):
            self.handler.run()
        log.info.assert_any_call("MBS build state is not 'ready', skipping")

    def fake_get_inheritance_data(self, tag):
        mock_inheritance = {
            'rhel-8.0-build': [
                {'name': 'tag1'},
                {'name': 'module-908765'},  # Remove inheritance from this one.
            ],
            'other-8.0-tag': [
                {'name': 'tag7'}
            ],
        }
        return mock_inheritance[tag]

    @mock.patch('ursa_major.handlers.add_tag.MailAPI')
    @mock.patch('ursa_major.koji_service.koji')
    @mock.patch('ursa_major.koji_service.KojiService.update_tag_inheritance')
    def test_update_module_koji_tag(self, update_tag_inheritance, koji, mail_api):
        self.handler.connect_mbs = mock.MagicMock()
        self.handler._mbs = mock.MagicMock()
        self.handler.mbs.get_modules.return_value = [
            {'koji_tag': 'module-123456'},
            {'koji_tag': 'module-908765'},
        ]

        build_url = 'http://www.example.com/job/test/1'

        # Not going to test kerberos auth. Just make sure login successfully.
        koji.get_profile_module.return_value.config.authtype = 'kerberos'
        session = koji.ClientSession.return_value
        session.getInheritanceData.side_effect = self.fake_get_inheritance_data

        def fake_update_tag_inheritance(tag, add_tags=None, remove_tags=None):
            if tag == "rhel-8.0-build":
                return (['module-123456'], ['module-908765'])
            elif tag == "other-8.0-tag":
                return (['module-123456'], [])

        update_tag_inheritance.side_effect = fake_update_tag_inheritance

        self.handler._mbs.get_module_mmd.return_value = make_mmd(
            'autotools', 'rhel-8.0', '20180101', '123456',
            {'platform': ['rhel-8.0']})

        ci_message = json.dumps(dict(
            id=123,
            state=MBS_BUILD_STATES['ready'],
            state_name='ready',
            koji_tag='module-123456'
        ))

        mock_env = {'CI_MESSAGE': ci_message,
                    'BUILD_URL': 'http://www.example.com/job/test/1'}

        self.set_args(send_mail=True)

        with mock.patch.dict(os.environ, mock_env):
            self.handler.run()

        update_tag_inheritance.assert_has_calls([
            mock.call('rhel-8.0-build',
                      add_tags=[{'name': 'module-123456', 'priority': 10}],
                      remove_tags=[{'name': 'module-908765'}]),
            mock.call('other-8.0-tag',
                      add_tags=[{'name': 'module-123456', 'priority': 30}],
                      remove_tags=[]),
        ], any_order=True)

        rhel_80_build_data = {
            'removed_tags': ['module-908765'],
            'tag': 'rhel-8.0-build',
            'added_tags': ['module-123456'],
            'module': {
                'build_id': 123,
                'nsvc': 'autotools:rhel-8.0:20180101:123456',
                'koji_tag': 'module-123456',
                'requires': {'platform': ['rhel-8.0']},
            },
            'build_url': build_url,
        }

        other_80_tag_data = {
            'removed_tags': [],
            'tag': 'other-8.0-tag',
            'added_tags': ['module-123456'],
            'module': {
                'build_id': 123,
                'nsvc': 'autotools:rhel-8.0:20180101:123456',
                'koji_tag': 'module-123456',
                'requires': {'platform': ['rhel-8.0']},
            },
            'build_url': build_url,
        }

        self.handler.mail_api.send_mail.assert_called()
        self.assertEqual(u'test', 'test')
        self.handler.mail_api.send_mail.assert_has_calls(
            [mock.call('add-tag.txt',
                       ['foo@example.com'],
                       data=rhel_80_build_data,
                       subject='Tag rhel-8.0-build is updated'),
             mock.call('add-tag.txt',
                       ['bar@example.com'],
                       data=other_80_tag_data,
                       subject='Tag other-8.0-tag is updated')],
            any_order=True)

    @mock.patch('ursa_major.handlers.add_tag.log')
    def test_skip_module(self, log):
        ci_message = json.dumps(dict(
            id=123,
            state=MBS_BUILD_STATES['ready'],
            state_name='ready',
            koji_tag='module-d243299e85d7e9aa'
        ))
        mock_mbs = mock.MagicMock()
        fake_mmd = make_mmd("testmodule", "master", "20180101", "1234567")
        mock_mbs.get_module_mmd.return_value = fake_mmd
        self.handler.connect_koji = mock.MagicMock()
        self.handler.connect_mbs = mock.MagicMock()
        self.handler._mbs = mock_mbs
        self.handler.manage_tag_inheritance = mock.MagicMock()

        with mock.patch.dict(os.environ, {'CI_MESSAGE': ci_message}):
            self.handler.run()

        self.handler.manage_tag_inheritance.assert_not_called()

        # If module represented by the CI_MESSAGE does not match any
        # criteria, just logging to skip this module. There are at
        # least two logging.info calls in addition to log other
        # messages.
        self.assertGreaterEqual(log.info.call_count, 2)

    def assert_get_module_config(self, fake_configs, expected_result):
        message_file = "testmodule_with_requires_ready_message.json"
        message = self.load_json_from_file(message_file)
        mock_mbs = mock.MagicMock()
        fake_requires = {"platform": ["el8"]}
        fake_mmd = make_mmd("testmodule", "rhel-8.0",
                            "20180409051516", "9e5fe74b",
                            requires=fake_requires,
                            buildrequires=fake_requires)
        mock_mbs.get_module_mmd.return_value = fake_mmd
        modinfo = ModuleInfo.from_mbs_message(mock_mbs, message)
        matched_config = AddTagHandler.get_module_config(fake_configs, modinfo)
        self.assertEqual(matched_config, expected_result)

    def test_get_module_config_match_requires(self):
        # ((fake_module_configs, expected result), ...)
        test_matrix = (
            (
                [
                    {
                        'name': 'testmodule', 'stream': 'rhel-8.0', 'priority': 10,
                        'requires': {'platform': 'el8'}
                    },
                    # This noisy config should not impact to choose the correct one.
                    {
                        'name': 'testmodule2', 'stream': 'rhel-8.0', 'priority': 20,
                    }
                ],
                {
                    'name': 'testmodule', 'stream': 'rhel-8.0', 'priority': 10,
                    'requires': {'platform': 'el8'}
                },
            ),
            (
                [
                    # buildrequires could be specified in config as well.
                    {
                        'name': 'testmodule', 'stream': 'rhel-8.0', 'priority': 10,
                        'buildrequires': {'platform': 'el8'}
                    },
                ],
                {
                    'name': 'testmodule', 'stream': 'rhel-8.0', 'priority': 10,
                    'buildrequires': {'platform': 'el8'}
                },
            ),
            (
                [
                    # Both requires and buildrequires are specified to match the
                    # module metadata.
                    {
                        'name': 'testmodule', 'stream': 'rhel-8.0', 'priority': 10,
                        'requires': {'platform': 'el8'},
                        'buildrequires': {'platform': 'el8'},
                    },
                ],
                {
                    'name': 'testmodule', 'stream': 'rhel-8.0', 'priority': 10,
                    'requires': {'platform': 'el8'},
                    'buildrequires': {'platform': 'el8'},
                },
            ),
            # Either requires or buildrequires is not included in module metadata.
            (
                [
                    {
                        'name': 'testmodule', 'stream': 'rhel-8.0', 'priority': 10,
                        'requires': {'platform': 'f30'},
                    },
                ],
                None,
            ),
            (
                [
                    # Both requires and buildrequires are specified to match the
                    # module metadata.
                    {
                        'name': 'testmodule', 'stream': 'rhel-8.0', 'priority': 10,
                        'buildrequires': {'platform': 'f30'},
                    },
                ],
                None,
            ),
            (
                [
                    {
                        'name': 'testmodule', 'stream': 'rhel-8.0', 'priority': 10,
                        'requires': {'platform': 'f30'},
                        'buildrequires': {'platform': 'f30'},
                    },
                ],
                None,
            ),
            # Either name or stream is not matched module metadata.
            (
                # Module name is not matched in this config
                [{'name': 'module-xxxx', 'stream': 'rhel-8.0', 'priority': 20}],
                None,
            ),
            (
                # Module stream is not matched in this config
                [{'name': 'testmodule', 'stream': '100', 'priority': 20}],
                None,
            ),
            # Module name and stream are matched
            (
                [
                    {'name': 'testmodule', 'stream': 'rhel-8.0', 'priority': 20},
                    {'name': 'testmodule2', 'stream': 'rhel-8.0', 'priority': 20},
                ],
                {'name': 'testmodule', 'stream': 'rhel-8.0', 'priority': 20},
            ),
        )
        for config, expected_result in test_matrix:
            self.assert_get_module_config(config, expected_result)

    @mock.patch('ursa_major.mbs.requests.get')
    def test_add_module_to_tag(self, requests_get):
        self.handler._mbs = MBS('http://mbs.example.com')
        self.handler._koji = mock.MagicMock()

        modules_file = 'mbs_modules_testmodule_all_in_one_page.json'
        json_data = self.load_json_from_file(modules_file)
        resp = MockResponse(json_data, 200)
        requests_get.return_value = resp

        message = self.load_json_from_file("testmodule_ready_message.json")
        mock_mbs = mock.MagicMock()
        fake_requires = {"platform": ["el8"]}
        fake_mmd = make_mmd("testmodule", "rhel-8.0",
                            "20180409051516", "9e5fe74b",
                            requires=fake_requires)
        mock_mbs.get_module_mmd.return_value = fake_mmd
        modinfo = ModuleInfo.from_mbs_message(mock_mbs, message)

        inheritance_tags = [
            {'name': 'tmp-build'},
            {'name': 'tmp-build-override'},
            {'name': 'testmodule-234'},
            {'name': 'module-358b0c179d25310c'}
        ]

        self.handler._koji.get_inheritance_data.return_value = \
            inheritance_tags

        self.handler._koji.update_tag_inheritance.return_value = (
            ['module-0c8d21d5c89ec363'], ['module-358b0c179d25310c'])

        module_config = ModuleConfig({
            'name': 'testmodule', 'stream': 'rhel-8.0',
            'priority': 10
        })
        added_tags, removed_tags = self.handler.add_module_to_tag(
            'test-tag', modinfo, module_config)

        self.handler.koji.update_tag_inheritance.assert_called_with(
            'test-tag',
            add_tags=[{'priority': 10, 'name': u'module-0c8d21d5c89ec363'}],
            remove_tags=[{'name': 'module-358b0c179d25310c'}]
        )
        expected_added_tags = ['module-0c8d21d5c89ec363']
        expected_removed_tags = ['module-358b0c179d25310c']

        self.assertEqual(set(added_tags), set(expected_added_tags))
        self.assertEqual(set(removed_tags), set(expected_removed_tags))
        self.handler.koji.regen_repos.assert_called()



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import math
import os
import json
try:
    import unittest2 as unittest
except ImportError:
    import unittest


from ursa_major.utils import Modulemd
from six.moves.configparser import ConfigParser

TESTS_DIR = os.path.dirname(os.path.realpath(__file__))
TEST_DATA_DIR = os.path.join(TESTS_DIR, 'test_data')

CONFIG_FILE = os.path.join(TEST_DATA_DIR, 'ursa-major-test.conf')
URSA_MAJOR_TEST_CONFIG = ConfigParser()
URSA_MAJOR_TEST_CONFIG.read(CONFIG_FILE)


def make_mmd(name, stream, version, context, requires=None, buildrequires=None):
    """Creates new Modulemd.Module instance

    :param name: module name
    :param stream: module stream
    :param version: module version
    :param context: module context
    :param requires: Dict of requires, example:
        {'platform': ['rhel-8.0'], 'python3': 'master'}
    :param buildrequires: Dict of build_requires, example:
        {'platform': 'rhel-8.0', 'bootstrap': ['rhel-8.0']}
    :rtype: Modulemd.Module instance
    """
    mmd = Modulemd.Module()
    mmd.set_name(name)
    mmd.set_stream(stream)
    mmd.set_version(int(version))
    mmd.set_context(context)
    # required options
    mmd.set_mdversion(2)
    mmd.set_summary("A test module in all its beautiful beauty.")
    description = ("This module demonstrates how to write simple "
                   "modulemd files And can be used for testing "
                   "the build and release pipeline.")
    mmd.set_description(description)
    licenses = Modulemd.SimpleSet()
    licenses.add("GPL")
    mmd.set_module_licenses(licenses)

    deps_list = []
    requires = requires or {}
    buildrequires = buildrequires or {}

    deps = Modulemd.Dependencies()
    for req_name, req_streams in requires.items():
        if not isinstance(req_streams, list):
            req_streams = [req_streams]
        deps.add_requires(req_name, req_streams)

    for req_name, req_streams in buildrequires.items():
        if not isinstance(req_streams, list):
            req_streams = [req_streams]
        deps.add_buildrequires(req_name, req_streams)
    deps_list.append(deps)
    mmd.set_dependencies(deps_list)

    return mmd


class MockResponse:
    def __init__(self, json_data, status_code):
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        return self.json_data


class UrsaMajorTestCase(unittest.TestCase):
    def load_json_from_file(self, filename):
        test_data_file = os.path.join(TEST_DATA_DIR, filename)
        with open(test_data_file, 'r') as f:
            return json.load(f)


class MockMBSBuildsData(object):
    def __init__(self, module_builds, per_page=2):
        self.module_builds = module_builds
        self.total = len(module_builds)
        self.per_page = per_page

    def _page_url(self, page_number):
        return (
            'https://mbs.example.com/module-build-service/1/module-builds/'
            '?per_page={}&page={}'.format(self.per_page, page_number)
        )

    def get(self, url, **kwargs):
        params = kwargs.get('params', {})
        q_page = params.get('page', None)
        q_per_page = params.get('per_page', None)

        page = q_page or 1

        if q_per_page is not None:
            self.per_page = q_per_page
        pages = int(math.ceil(float(self.total)/float(self.per_page)))

        start = self.per_page * (page - 1)
        stop = self.per_page * page
        modules = self.module_builds[start:stop]

        next_page_url = None
        if pages > page:
            next_page_url = self._page_url(page + 1)

        prev_page_url = None
        if page > 1:
            prev_page_url = self._page_url(page - 1)

        json_data = {
            'items': modules,
            'meta': {
                'first': self._page_url(1),
                'last': self._page_url(pages),
                'next': next_page_url,
                'page': page,
                'pages': pages,
                'per_page': self.per_page,
                'prev': prev_page_url,
                'total': self.total
            }
        }
        return MockResponse(json_data, 200)



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

from ursa_major.handlers.base import BaseHandler
from ursa_major import ModuleConfig, ModuleConfigCmp, MBS_BUILD_STATES
from ursa_major.logger import log


class AddModuleHandler(BaseHandler):
    """Handler to add a module to config file and add its koji tag to inheritance data"""

    def run(self):
        self.dry_run = self.args.dry_run
        self.debug = self.args.debug

        self.force = self.args.force

        self.tag_config_file = self.args.tag_config_file
        self.load_tag_config()

        self.wait_regen_repo = self.args.wait_regen_repo

        # the koji tag we operate against
        self.op_tag = self.args.tag

        tags_to_remove = []
        tag_to_add = None
        inheritance_updated = False

        self.connect_mbs()
        self.connect_koji(dry_run=self.dry_run)
        self.koji.login()

        # By default, if necessary, should update specified tag's
        # inheritance data to add or edit the latest module's tag,
        # and the tag config file should be updated as well.
        to_update_inheritance = True
        to_update_config = True

        if self.args.update_inheritance_only:
            to_update_config = False
            log.warning("Will not update config file because --update-inheritance-only "
                        "is specified, even if the specified module config does not "
                        "exist in tag config file.")
        if self.args.update_config_only:
            to_update_inheritance = False
            log.warning("Will not update tag %s's inheritance data because --update-config-only "
                        "is specified, even if the latest module build's tag does not exists in "
                        "inheritance with specified config", self.op_tag)

        if self.args.module_requires:
            module_requires = dict(self.args.module_requires)
        else:
            module_requires = None

        if self.args.module_buildrequires:
            module_buildrequires = dict(self.args.module_buildrequires)
        else:
            module_buildrequires = None

        input_module_config = ModuleConfig(
            module_config=dict(
                name=self.args.module_name,
                stream=self.args.module_stream,
                priority=self.args.priority,
                requires=module_requires,
                buildrequires=module_buildrequires,
            ))

        # existing tags of specified module from cmdline
        existing_input_module_tags = \
            self.find_module_tags_from_koji_inheritance(self.op_tag, input_module_config)

        if existing_input_module_tags and to_update_inheritance:
            log.info("There are tags exist for the module specified in command line: %s",
                     existing_input_module_tags)
            # Add these existing tags to list, but we may remove the one of
            # latest_module_tag later
            tags_to_remove.extend(existing_input_module_tags)

        # Find out the latest module tag of specified module in MBS
        latest_module_tag = self.find_latest_module_tag(input_module_config)
        if latest_module_tag:
            log.info("Found koji tag %s of the latest module in MBS with module config '%s'",
                     latest_module_tag, input_module_config)

            # in all possible cases, this is the tag we're going to add if needed
            tag_to_add = latest_module_tag

        else:
            log.warning("Can't find any built module in MBS with module config '%s', "
                        "we will not add any new module tag to inheritance data",
                        input_module_config)
            if not self.force:
                log.warning("Will not update inheritance data nor tag config file, consider "
                            "to run with --force to allow updating")
                to_update_inheritance = False
                to_update_config = False

        # find any module exist in config with the same name and stream
        stock_module_config = self.find_module_config(self.op_tag, input_module_config)
        if stock_module_config:
            log.info("Found module config in tag configuration that has same "
                     "name and stream of the specified module: %s",
                     stock_module_config)
        else:
            log.info('No corresponding module config found in tag config file.')

        config_diff = input_module_config.compare(stock_module_config)

        # In all of the cases, we just need to find out the tags to be removed,
        # and then check:
        # 1. if tag_to_add is None:
        #    remove all tags_to_remove
        # 2. if tag_to_add is not None and in tags_to_remove:
        #    (a). if tag's specified priority has same value with the value
        #         in current inheritance data, remove tag_to_add from
        #         tags_to_remove, and then remove tags in tags_to_remove
        #         TODO: or maybe we should do nothing in this case?
        #    (b). if tag's specified priority is different than the value
        #         in current inheritance data, remove tag_to_add from
        #         tags_to_remove, then add tag_to_add (with the specified
        #         priority) and remove tags in tags_to_remove at the same time
        # 3. if tag_to_add is not None and not in tags_to_remove:
        #    remove tags in tags_to_remove and add tag_to_add

        if config_diff == ModuleConfigCmp.NEW:
            log.info("The specified module does not exist in tag config file "
                     "under tag '%s'", self.op_tag)
            # this is a new module in config, check whether the priority exist
            existing_priorities = self.get_module_priorities_in_config(self.op_tag)
            if input_module_config.priority in existing_priorities:
                raise RuntimeError("The specified priority '{}' conflicts with the "
                                   "existing priorities under tag '{}'".format(
                                       input_module_config.priority, self.op_tag))
            # in this case, the possible tags need to remove is the existing_input_tags,
            # which has been added to tags_to_remove before

        elif config_diff == ModuleConfigCmp.DIFFERENT:
            # Module in config changes, it can be a change in requires,
            # buildrequires or priority.

            log.info("A module with same name and stream exists in tag config file "
                     "under tag '%s', but has different configuration", self.op_tag)

            # Check whether the specified priority conflicts with other modules
            existing_priorities = self.get_module_priorities_in_config(self.op_tag)
            if (input_module_config.priority != stock_module_config.priority and
                    input_module_config.priority in existing_priorities):
                raise RuntimeError("The specified priority '{}' conflicts with the "
                                   "existing priorities under tag '{}'".format(
                                       input_module_config.priority, self.op_tag))

            if to_update_inheritance:
                if (input_module_config.requires != stock_module_config.requires or
                        input_module_config.buildrequires != stock_module_config.buildrequires):
                    # module requires changed, we need to remove old tags
                    existing_stock_tags = self.find_module_tags_from_koji_inheritance(
                        self.op_tag, stock_module_config)
                    if existing_stock_tags:
                        log.info("Module config is changed, there are tags of old module "
                                 "exist in inheritance data:\n%s", existing_stock_tags)
                        # Add these tags to remove list
                        tags_to_remove.extend(existing_stock_tags)
                        # we may adding duplicated tags?
                        tags_to_remove = list(set(tags_to_remove))

                # else: module priority changed, that will be handled later,
                # nothing to do here

        elif config_diff == ModuleConfigCmp.SAME:
            log.info("An module exist in tag config file under tag '%s' with the "
                     "same configuration, we will not update anything", self.op_tag)
            to_update_inheritance = False
            to_update_config = False

        if to_update_inheritance:
            if tag_to_add is None and tags_to_remove:
                remove_tags = [{'name': t} for t in tags_to_remove]
                log.info("No module tag to add and going to remove tags:\n%s",
                         tags_to_remove)
                self.koji.update_tag_inheritance(self.op_tag,
                                                 remove_tags=remove_tags)
                inheritance_updated = True

            if tag_to_add is not None:
                if tag_to_add in tags_to_remove:
                    tags_to_remove.remove(tag_to_add)
                    remove_tags = [{'name': t} for t in tags_to_remove]
                    if remove_tags:
                        log.info("Going to remove old tags: %s", tags_to_remove)

                    cur_inheritance = self.koji.get_inheritance_data(self.op_tag)
                    old_tag_data = [d for d in cur_inheritance if d['name'] == tag_to_add].pop()
                    if old_tag_data.get('priority') == input_module_config.priority:
                        # don't need to add the tag since it already exist with
                        # the same priority as specified
                        log.info("The latest module tag %s already exists with same priority "
                                 "as specified, will not add any tag to %s's inheritance data",
                                 tag_to_add, self.op_tag)
                        if remove_tags:
                            self.koji.update_tag_inheritance(self.op_tag,
                                                             remove_tags=remove_tags)
                            inheritance_updated = True
                        # remove list is empty, we don't need to remove any tag
                    else:
                        # latest module tag is in inheritance data, but priority is different,
                        # we need to update the priority
                        log.info("The latest module tag %s exists but with different priority, "
                                 "will update its priority in inheritance", tag_to_add)
                        edit_tags = [{'name': tag_to_add,
                                      'priority': input_module_config.priority}]
                        self.koji.update_tag_inheritance(self.op_tag,
                                                         remove_tags=remove_tags,
                                                         edit_tags=edit_tags)
                        inheritance_updated = True

                # tag_to_add is not in tags_to_remove
                else:
                    add_tags = [{'name': tag_to_add,
                                 'priority': input_module_config.priority}]
                    remove_tags = [{'name': t} for t in tags_to_remove]
                    log.info("Going to add latest module tag: %s", tag_to_add)
                    log.info("Going to remove old tags: %r", tags_to_remove)
                    self.koji.update_tag_inheritance(self.op_tag,
                                                     add_tags=add_tags,
                                                     remove_tags=remove_tags)
                    inheritance_updated = True

        if inheritance_updated:
            if self.koji.is_build_tag(self.op_tag):
                build_tags = [self.op_tag]
            else:
                build_tags = self.koji.find_build_tags(self.op_tag)

            if not build_tags:
                log.warning("No build tag found for %s, not regen any repo", self.op_tag)
            else:
                log.info("Found build tags for %s: %r", self.op_tag, build_tags)
                self.koji.regen_repos(build_tags, wait=self.wait_regen_repo)

        if to_update_config:
            self.add_module_to_config(self.op_tag, input_module_config)

        # Log out from Koji session before leave
        self.koji.logout()

    def get_module_priorities_in_config(self, tag):
        """
        Get all priorities of modules under specified tag in tag config file.

        :param tag: tag name
        """
        config = self.tag_config.get(tag, {})
        modules = config.get('modules', [])
        return [m['priority'] for m in modules]

    def find_latest_module_tag(self, module_config):
        """
        Find the latest ready module which matches the module_config
        in MBS, return its koji_tag, or return None if not found.

        :param module_config: an object of `ModuleConfig`
        :return: tag name or None
        """
        modules = self.mbs.get_modules(
            buildrequires=module_config.buildrequires,
            requires=module_config.requires,
            name=module_config.name,
            stream=module_config.stream,
            state=MBS_BUILD_STATES['ready'],
            page=1)  # we only need the default first page items
        if modules:
            return modules[0]['koji_tag']
        else:
            return None

    def find_module_config(self, tag, module_config):
        """
        Find matched module config under specified tag by name and
        stream in config file

        :param tag: tag name
        :return: object of `ModuleConfig` if found. None otherwise.
        :rtype: ModuleConfig
        """
        tag_config = self.tag_config
        if tag not in tag_config:
            return None
        modules = tag_config[tag].get('modules', [])
        for config in modules:
            if config['name'] == module_config.name and \
                    config['stream'] == module_config.stream:
                return ModuleConfig(config)
        return None

    def add_module_to_config(self, tag, module_config):
        """
        Add the module config to tag config file

        :param tag: tag name, module config will be added under this tag
        :param module_config: an object of `ModuleConfig`
        """
        tag_config = self.tag_config
        config = tag_config.setdefault(tag, {})
        modules = config.setdefault('modules', [])
        module_config_dict = dict(name=module_config.name,
                                  stream=module_config.stream,
                                  priority=module_config.priority)
        if module_config.requires:
            module_config_dict['requires'] = module_config.requires
        if module_config.buildrequires:
            module_config_dict['buildrequires'] = module_config.buildrequires

        modules_with_same_name_stream = [m for m in modules if
                                         m['name'] == module_config.name and
                                         m['stream'] == module_config.stream]

        if len(modules_with_same_name_stream) > 1:
            raise RuntimeError("Unexpected error, found multiple modules in config "
                               "file with same name (%s) and stream (%s)".format(
                                   module_config.name, module_config.stream))

        if len(modules_with_same_name_stream) == 0:
            log.info("Adding module config to tag %s in tag config file:\n%s",
                     tag, module_config_dict)
            modules.append(module_config_dict)

        if len(modules_with_same_name_stream) == 1:
            log.info("Found one existing module config with same name and stream, "
                     "update it with:\n%s", module_config_dict)
            modules.remove(modules_with_same_name_stream[0])
            modules.append(module_config_dict)

        self.write_tag_config()



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import json
import mock
import os
import six
import shutil
import tempfile
try:
    import unittest2 as unittest
except ImportError:
    import unittest

from argparse import Namespace
from ursa_major import MBS_BUILD_STATES
from ursa_major.handlers.update_tag import UpdateTagHandler


class TestUpdateTagHandler(unittest.TestCase):
    def setUp(self):
        config = mock.MagicMock()
        self.handler = UpdateTagHandler(config)

        self.handler.connect_koji = mock.MagicMock()
        self.handler._koji = mock.MagicMock()

        self.handler.connect_mbs = mock.MagicMock()
        self.handler._mbs = mock.MagicMock()

        self.tmpdir = tempfile.mkdtemp(suffix='_ursa_major_test')
        self.tag_config_file = os.path.join(self.tmpdir, 'default.json')
        tag_config = {
            'example-tag': {
                'owners': ['foo@example.com'],
                'modules': [
                    {'name': 'testmodule', 'stream': 'f30', 'priority': 10,
                     'requires': {'platform': 'f30'}},
                ]
            },
            'empty-tag': {
                'owners': ['foo@example.com'],
                'modules': []
            }
        }

        with open(self.tag_config_file, 'w') as f:
            json.dump(tag_config, f)

        self.set_args()

    def tearDown(self):
        try:
            shutil.rmtree(self.tmpdir)
        except:  # noqa
            pass

    def set_args(self, **kwargs):
        kwargs.setdefault('tag_config_file', self.tag_config_file)
        kwargs.setdefault('wait_regen_repo', False)
        kwargs.setdefault('dry_run', False)
        kwargs.setdefault('debug', False)
        kwargs.setdefault('tag', 'example-tag')
        args = Namespace()
        for k, v in kwargs.items():
            setattr(args, k, v)
        self.handler.set_args(args)

    def mock_get_tag(self, tag):
        taginfo = {
            "example-tag": {
                "name": "example-tag",
                "id": 123
            },
            "module-testmodule-f30-20180101-abc123": {
                "name": "module-testmodule-f30-20180101-abc123",
                "id": 10034
            },
            "module-testmodule-f30-20161212-000000": {
                "name": "module-testmodule-f30-20161212-000000",
                "id": 8080
            },
        }
        return taginfo.get(tag, {})

    def test_terminate_if_tag_not_in_config(self):
        self.set_args(tag='non-exist-tag')
        with six.assertRaisesRegex(self, RuntimeError, 'is not found in tag config file'):
            self.handler.run()

    def test_terminate_if_tag_not_in_koji(self):
        self.handler.koji.get_tag.return_value = None
        with six.assertRaisesRegex(self, RuntimeError, 'is not found in koji'):
            self.handler.run()

    @mock.patch('ursa_major.handlers.update_tag.log')
    def test_terminate_if_tag_has_no_module_in_config(self, log):
        self.set_args(tag='empty-tag')
        self.handler.run()
        log.warning.assert_any_call(
            "No module specified for tag '%s' in tag config file", "empty-tag")

    @mock.patch('ursa_major.handlers.update_tag.log')
    def test_skip_if_no_ready_build_found(self, log):
        self.handler.mbs.get_modules.return_value = []

        self.handler.run()
        log.warning.assert_has_calls([
            mock.call("There is no ready build found for %s, skipping", mock.ANY)
        ])

    @mock.patch('ursa_major.handlers.update_tag.log')
    def test_skip_if_tag_in_inheritance_with_same_priority(self, log):
        self.handler.mbs.get_modules.return_value = [
            {'koji_tag': 'module-testmodule-f30-20180101-abc123'}
        ]
        mock_inheritance = [
            {'name': 'module-testmodule-f30-20180101-abc123',
             'priority': 10}
        ]
        self.handler.koji.get_inheritance_data.return_value = mock_inheritance
        self.handler.run()
        log.info.assert_has_calls([
            mock.call("Tag '%s' is in inheritance data of %s with same priority, skipping",
                      "module-testmodule-f30-20180101-abc123", "example-tag")
        ])

    @mock.patch('ursa_major.handlers.update_tag.log')
    def test_tag_in_inheritance_with_different_priority(self, log):
        self.handler.mbs.get_modules.return_value = [
            {'koji_tag': 'module-testmodule-f30-20180101-abc123'}
        ]
        mock_inheritance = [
            {'name': 'module-testmodule-f30-20180101-abc123',
             'priority': 50}
        ]
        self.handler.koji.get_inheritance_data.return_value = mock_inheritance
        self.handler.run()
        self.handler.koji.set_inheritance_data.assert_has_calls([
            mock.call('example-tag',
                      [{'priority': 10, 'name': 'module-testmodule-f30-20180101-abc123'}])
        ])
        self.handler.koji.regen_repo.assert_has_calls([
            mock.call('example-tag', wait=False)
        ])

    @mock.patch('ursa_major.handlers.update_tag.log')
    def test_tag_not_in_inheritance_and_no_old_tag_found(self, log):
        self.handler.mbs.get_modules.return_value = [
            {'koji_tag': 'module-testmodule-f30-20180101-abc123'}
        ]

        self.handler.koji.get_tag.side_effect = self.mock_get_tag

        self.handler.koji.get_inheritance_data.return_value = []
        self.handler.run()
        inheritance_data = [
            {'intransitive': False,
             'name': 'module-testmodule-f30-20180101-abc123',
             'parent_id': 10034,
             'priority': 10,
             'maxdepth': None,
             'noconfig': False,
             'pkg_filter': '',
             'child_id': 123}
        ]
        self.handler.koji.set_inheritance_data.assert_has_calls([
            mock.call('example-tag', inheritance_data)
        ])
        self.handler.koji.regen_repo.assert_has_calls([
            mock.call('example-tag', wait=False)
        ])

    def test_tag_not_in_inheritance_and_old_tag_exist(self):
        mock_inheritance = [
            {'name': 'module-testmodule-f30-20161212-000000',
             'priority': 10}
        ]
        self.handler.koji.get_inheritance_data.return_value = mock_inheritance
        self.handler.koji.get_tag.side_effect = self.mock_get_tag

        def mock_get_modules(**kwargs):
            state = kwargs.get('state', None)
            if isinstance(state, list) and MBS_BUILD_STATES['garbage'] in state:
                # return all tags
                return [
                    {'koji_tag': 'module-testmodule-f30-20161212-000000'},
                    {'koji_tag': 'module-testmodule-f30-20180101-abc123'},
                ]
            else:
                # only return the latest one
                return [{'koji_tag': 'module-testmodule-f30-20180101-abc123'}]

        self.handler.mbs.get_modules.side_effect = mock_get_modules

        self.handler.run()

        call_args = self.handler.koji.set_inheritance_data.call_args.args
        self.assertEqual('example-tag', call_args[0])
        tag_to_remove = [t['name'] for t in call_args[1] if t.get('delete link', False) is True]
        tag_to_add = [t['name'] for t in call_args[1] if t.get('delete link', False) is False]
        self.assertTrue("module-testmodule-f30-20161212-000000" in tag_to_remove)
        self.assertTrue("module-testmodule-f30-20180101-abc123" in tag_to_add)
        self.handler.koji.regen_repo.assert_has_calls([
            mock.call('example-tag', wait=False)
        ])



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

from tests import UrsaMajorTestCase
from ursa_major.utils import mmd_has_requires
from tests import make_mmd


class TestMmdHasRequires(UrsaMajorTestCase):
    def test_mmd_has_neg_requires(self):
        requires = {'platform': 'f26'}
        mmd = make_mmd('testmodule', 'master', '123', '00000000', requires)
        check_requires = {'platform': '-f26'}
        result = mmd_has_requires(mmd, check_requires)
        self.assertFalse(result)

    def test_mmd_missing_requires(self):
        requires = {'platform': 'f28'}
        mmd = make_mmd('testmodule', 'master', '123', '00000000', requires)
        check_requires = {'platform': 'f28', 'python3': 'master'}
        result = mmd_has_requires(mmd, check_requires)
        self.assertFalse(result)

    def test_mmd_not_has_neg_requires(self):
        requires = {'platform': 'f28', 'python3': 'master'}
        mmd = make_mmd('testmodule', 'master', '123', '00000000', requires)
        check_requires = {'platform': '-f26', 'python3': 'master'}
        result = mmd_has_requires(mmd, check_requires)
        self.assertTrue(result)

    def test_mmd_does_not_has_pos_requires(self):
        requires = {'platform': 'f28', 'python3': 'master'}
        mmd = make_mmd('testmodule', 'master', '123', '00000000', requires)
        check_requires = {'platform': 'f28', 'python3': 'f28'}
        result = mmd_has_requires(mmd, check_requires)
        self.assertFalse(result)

    def test_mmd_has_more_requires_than_checks(self):
        requires = {'platform': 'f28', 'python3': 'master'}
        mmd = make_mmd('testmodule', 'master', '123', '00000000', requires)
        check_requires = {'platform': 'f28'}
        result = mmd_has_requires(mmd, check_requires)
        self.assertTrue(result)

    def test_mmd_has_more_require_streams_than_checks(self):
        requires = {'platform': ['f28', 'f29'], 'python3': 'master'}
        mmd = make_mmd('testmodule', 'master', '123', '00000000', requires)
        check_requires = {'platform': 'f28'}
        result = mmd_has_requires(mmd, check_requires)
        self.assertTrue(result)

    def test_mmd_has_a_neg_stream_in_streams_list(self):
        requires = {'platform': ['f28', 'f29'], 'python3': 'master'}
        mmd = make_mmd('testmodule', 'master', '123', '00000000', requires)
        check_requires = {'platform': '-f28'}
        result = mmd_has_requires(mmd, check_requires)
        self.assertFalse(result)

    def test_mmd_has_a_neg_stream_in_multi_neg_streams(self):
        requires = {'platform': ['f27'], 'python3': 'master'}
        mmd = make_mmd('testmodule', 'master', '123', '00000000', requires)
        check_requires = {'platform': ['-f26', '-f27']}
        result = mmd_has_requires(mmd, check_requires)
        self.assertFalse(result)

    def test_check_with_empty_requires(self):
        requires = {'platform': ['f27'], 'python3': 'master'}
        mmd = make_mmd('testmodule', 'master', '123', '00000000', requires)
        result = mmd_has_requires(mmd, {})
        self.assertTrue(result)



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import copy
import json
import traceback

from ursa_major import ModuleConfig
from ursa_major.handlers.base import BaseHandler
from ursa_major.logger import log
from ursa_major.mail import MailAPI
from ursa_major.utils import get_env_var, mmd_has_requires, mmd_has_buildrequires


class ModuleInfo(object):
    def __init__(self):
        fields = ('build_id', 'koji_tag', 'mmd',
                  'name', 'stream', 'version', 'context',
                  'requires', 'buildrequires')
        for field in fields:
            setattr(self, field, None)

    @classmethod
    def from_mbs_message(cls, mbs, message):
        """
        Create an instance from mbs state change message.

        :param message: message in dict format
        :param mbs: MBS instance
        :return: an instance of ModuleInfo
        """
        modinfo = cls()
        modinfo.koji_tag = message['koji_tag']
        modinfo.build_id = message['id']

        mmd = mbs.get_module_mmd(modinfo.build_id)
        modinfo.mmd = mmd
        modinfo.name = mmd.peek_name()
        modinfo.stream = mmd.peek_stream()
        modinfo.version = str(mmd.peek_version())
        modinfo.context = mmd.peek_context()

        deps_list = mmd.peek_dependencies()
        requires = deps_list[0].peek_requires() if deps_list else {}
        buildrequires = deps_list[0].peek_buildrequires() if deps_list else {}
        modinfo.requires = {name: stream.dup() for
                            name, stream in requires.items()}
        modinfo.buildrequires = {name: stream.dup() for
                                 name, stream in buildrequires.items()}

        return modinfo

    @property
    def nsvc(self):
        return "{}:{}:{}:{}".format(
               self.name, self.stream, self.version, self.context)

    @property
    def short_info(self):
        return {
            'build_id': self.build_id,
            'koji_tag': self.koji_tag,
            'nsvc': self.nsvc,
            'requires': self.requires,
        }


class AddTagHandler(BaseHandler):
    def __init__(self, *args, **kwargs):
        super(AddTagHandler, self).__init__(*args, **kwargs)
        self.send_mail = False
        self.wait_regen_repo = False

    @staticmethod
    def get_module_config(module_configs, modinfo):
        """
        Get the module config from configs which matches the module represented
        by the modinfo.

        :param module_configs: a list of module config
        :type module_configs: list[dict]
        :param modinfo: instance of ModuleInfo
        :type modinfo: :class:`ModuleInfo`
        :return: the matched module config. None is returned if no module
            config matches the module metadata.
        :rtype: dict
        :raises RuntimeError: if more than one module configs match the
            specified module metadata.
        """
        matched = []
        for config in module_configs:
            name = config['name']
            stream = config['stream']
            if not (name == modinfo.name and stream == modinfo.stream):
                continue
            dep_requires = config.get('buildrequires')
            if dep_requires and not mmd_has_buildrequires(modinfo.mmd, dep_requires):
                continue
            requires = config.get('requires', {})
            if requires and not mmd_has_requires(modinfo.mmd, requires):
                continue
            matched.append(config)

        if len(matched) == 0:
            return None
        if len(matched) > 1:
            error = ("Multiple config items match with module <{}>, unable to "
                     "determine which item we should use: {!s}".format(
                         modinfo.nsvc, matched))
            log.error(error)
            raise RuntimeError(error)
        return matched[0]

    def add_module_to_tag(self, tag, modinfo, module_config):
        """
        Add koji tag of module represent by modinfo to specified tag's
        inheritance data, config options are in module_config.

        Note: while adding module tag to inheritance, we remove old tags
        of the module as well. We do it in one call to avoid race
        condition.

        :param tag: tag name
        :param modinfo: ModuleInfo instance
        :param module_config: a module config object.
        :type module_config: :class:`ModuleConfig`
        :return: a tuple of (added_tags, removed_tags)
        :rtype: tuple[list[str], list[str]]
        """
        old_tags = self.find_module_tags_from_koji_inheritance(tag, module_config)

        # if this is a tag already exists, we will try to update it
        if modinfo.koji_tag in old_tags:
            old_tags.remove(modinfo.koji_tag)

        add_tag_info = {
            'name': modinfo.koji_tag,
            'priority': module_config.priority
        }

        self.koji.login()
        added_tags, removed_tags = self.koji.update_tag_inheritance(
            tag,
            add_tags=[add_tag_info],
            remove_tags=[{'name': name} for name in old_tags]
        )

        # regen repo for build tags only when inheritance data is updated
        if added_tags or removed_tags:
            if self.koji.is_build_tag(tag):
                build_tags = [tag]
            else:
                build_tags = self.koji.find_build_tags(tag)

            if not build_tags:
                log.warning("No build tag found for %s, not regen any repo", tag)
            else:
                log.info("Found build tags for '%s': %r", tag, build_tags)
                self.koji.regen_repos(build_tags, wait=self.wait_regen_repo)

        return added_tags, removed_tags

    def run(self):
        try:
            ci_message_string = get_env_var(self.args.module_from_env,
                                            raise_if_not_exist=True)
            ci_message = json.loads(ci_message_string)
            log.info('Received MBS state change message:\n%s',
                     json.dumps(ci_message, indent=2, sort_keys=True))

        except Exception as e:
            error = "Fail to load '{!s}' from os environ: {!s}".format(self.args.module_from_env, e)
            log.error(error)
            raise RuntimeError(error)

        if ci_message.get('state_name', None) != 'ready':
            log.info("MBS build state is not 'ready', skipping")
            return

        self.dry_run = self.args.dry_run
        self.send_mail = self.args.send_mail
        self.tag_config_file = self.args.tag_config_file
        self.wait_regen_repo = self.args.wait_regen_repo

        self.load_tag_config()

        if self.send_mail:
            if not self.config.has_section('mail'):
                log.error("Sending mail is enabled, but config doesn't contain mail settings")
                raise RuntimeError("Missing mail settings")
            mail_config = dict(self.config.items('mail'))
            self.mail_api = MailAPI(mail_config)
            self.mail_data = {'build_url': get_env_var('BUILD_URL')}

        self.connect_koji(dry_run=self.dry_run)
        self.connect_mbs()

        modinfo = ModuleInfo.from_mbs_message(self.mbs, ci_message)
        if self.send_mail:
            self.mail_data['module'] = modinfo.short_info

        errors = []
        for tag, tag_config in self.tag_config.items():
            tag_owners = tag_config.get('owners', [])
            if not tag_owners:
                log.warning("Sending mail is enabled, but no owners set for tag %s, "
                            "will not send mail", tag)

            conf_modules = tag_config.get('modules', [])
            try:
                module_config = self.get_module_config(conf_modules, modinfo)
                if module_config is None:
                    log.info("Module %s is not listed in tag %s, skipping this tag.",
                             modinfo.nsvc, tag)
                    continue

                # the module we're handling matches config of this tag
                added_tags, removed_tags = self.add_module_to_tag(
                    tag, modinfo, ModuleConfig(module_config))

                if self.dry_run or not tag_owners:
                    continue

                if (added_tags or removed_tags) and self.send_mail and tag_owners:
                    # send tag update success mail
                    mail_subject = 'Tag {} is updated'.format(tag)
                    content = copy.deepcopy(self.mail_data)
                    content['tag'] = tag
                    content['added_tags'] = added_tags
                    content['removed_tags'] = removed_tags
                    log.info("Sending mail to %r with content:\n%r", str(tag_owners), content)
                    self.mail_api.send_mail('add-tag.txt', tag_owners,
                                            subject=mail_subject, data=content)

            except Exception as e:
                error = ("Error while adding module <{!s}> to tag <{}>: "
                         "{!s}".format(modinfo.nsvc, tag, e))
                errors.append(error)
                log.error(error)

                if self.dry_run or not tag_owners:
                    continue

                if self.send_mail and tag_owners:
                    mail_subject = "Failed to update tag {}".format(tag)
                    content = copy.deepcopy(self.mail_data)
                    content['tag'] = tag
                    content['traceback'] = traceback.format_exc()
                    log.info("Sending mail to %r with data:\n%r", tag_owners, content)
                    self.mail_api.send_mail('add-tag-error.txt', tag_owners,
                                            subject=mail_subject, data=content)

        if errors:
            raise RuntimeError("Errors: {!s}".format(errors))



# -*- coding: utf-8 -*-
# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import json
import os
import tempfile

from mock import patch, Mock
from ursa_major.cli import main
from tests import TEST_DATA_DIR, MockMBSBuildsData, make_mmd


class TestRemoveModule:
    """Test CLI command remove-module"""

    def setup_method(self, method):
        fd, self.tag_config_file = tempfile.mkstemp()
        os.close(fd)

    def teardown_method(self, method):
        os.unlink(self.tag_config_file)

    def write_tag_config_file(self, content):
        with open(self.tag_config_file, 'w') as f:
            json.dump(content, f)

    def run_cli(self, options):
        config_file = os.path.join(TEST_DATA_DIR, 'ursa-major-test.conf')
        cli_cmd = ['ursa-major', 'remove-module', '--config', config_file] + options

        with patch('sys.argv', new=cli_cmd):
            main()

    @patch('koji.ClientSession')
    @patch('ursa_major.mbs.requests.get')
    def test_module_koji_tag_isnt_present_in_tag_inheritance(self, get, ClientSession):
        session = ClientSession.return_value
        session.getInheritanceData.return_value = [
            {'name': 'module-ant-123'},
            {'name': 'module-java-456'},
        ]

        get.return_value = Mock(status_code=200)
        get.return_value.json.return_value = {
            'items': [
                {
                    # This koji_tag is not present in the tag inheritance, so
                    # no tag is going to be removed.
                    'koji_tag': 'module-python-3.6',
                    'modulemd': make_mmd(name='mariadb', stream='10.4',
                                         version='3020190313091759', context='a5b0195c',
                                         requires={'platform': 'f30'},
                                         buildrequires={'platform': 'f30'}).dumps(),
                },
            ],
            'meta': {
                'prev': None,
                'last': None,
                'next': None,
                'total': 1,
                'per_page': 10,
                'first': 'https://mbs.example.com/module-build-service/1/module-builds/'
                         '?per_page=10&page=1&name=testmodule',
                'pages': 1,
                'page': 1
            }
        }

        # Write empty tag config file because we do not want to test if the tag
        # config file is written.
        self.write_tag_config_file({})

        self.run_cli([
            '--name', 'python', '--stream', '3.6', '--tag', 'f30-build',
            '--tag-config-file', self.tag_config_file
        ])

        session.setInheritanceData.assert_not_called()

    @patch('koji.ClientSession')
    @patch('ursa_major.mbs.requests.get')
    def remove_module(self, not_update_file, get, ClientSession):
        session = ClientSession.return_value
        session.getInheritanceData.return_value = [
            {'parent_id': 20,
             'name': 'module-ant-1.10-20181122140939-819b5873'},
            {'parent_id': 100,
             'name': 'module-mariadb-10.4-3020190304180835-a5b0195c'},
        ]
        session.getTag.return_value = {
            'id': 100,
            'name': 'module-mariadb-10.4-3020190304180835-a5b0195c',
        }

        mock_builds = [
            {
                'koji_tag': 'module-mariadb-10.4-3020190313091759-a5b0195c',
                'modulemd': make_mmd(name='mariadb', stream='10.4',
                                     version='3020190313091759', context='a5b0195c',
                                     requires={'platform': 'f30'},
                                     buildrequires={'platform': 'f30'}).dumps(),
            },
            {
                # This koji_tag is present in the inheritance, so it should be removed.
                'koji_tag': 'module-mariadb-10.4-3020190304180835-a5b0195c',
                'modulemd': make_mmd(name='mariadb', stream='10.4',
                                     version='3020190304180835', context='a5b0195c',
                                     requires={'platform': 'f30'},
                                     buildrequires={'platform': 'f30'}).dumps(),
            }
        ]
        mock_mbs_builds_data = MockMBSBuildsData(mock_builds)
        get.side_effect = mock_mbs_builds_data.get

        cmd = [
            '--name', 'mariadb', '--stream', '10.4',
            '--require', 'platform:f30', '--buildrequire', 'platform:f30',
            '--tag', 'f30-build',
            '--tag-config-file', self.tag_config_file
        ]
        if not_update_file:
            cmd.append('--not-update-config')

        self.run_cli(cmd)

        data = session.getInheritanceData.return_value[:]
        data[-1]['delete link'] = True
        session.setInheritanceData.assert_called_once_with('f30-build', data)

    def test_remove_module_and_remove_module_config_from_tag_configs(self):
        fake_tag_configs = {
            'f30-build': {
                'modules': [
                    {
                        'name': 'mariadb',
                        'stream': '10.4',
                        'requires': {'platform': 'f30'},
                        'buildrequires': {'platform': 'f30'},
                    }
                ],
                'owners': ['someone@example.com']
            }
        }
        self.write_tag_config_file(fake_tag_configs)

        self.remove_module(False)

        with open(self.tag_config_file, 'r') as f:
            tag_configs = json.load(f)

        mariadb_module_config = fake_tag_configs['f30-build']['modules'][0]
        assert mariadb_module_config not in tag_configs['f30-build']['modules']

    def test_remove_module_and_not_remove_module_config_from_tag_configs(self):
        fake_tag_configs = {
            'f30-build': {
                'modules': [
                    {
                        'name': 'mariadb',
                        'stream': '10.4',
                        'requires': {'platform': 'f30'},
                        'buildrequires': {'platform': 'f30'},
                    }
                ],
                'owners': ['someone@example.com']
            }
        }
        self.write_tag_config_file(fake_tag_configs)

        self.remove_module(True)

        with open(self.tag_config_file, 'r') as f:
            tag_configs = json.load(f)

        mariadb_module_config = fake_tag_configs['f30-build']['modules'][0]
        assert mariadb_module_config in tag_configs['f30-build']['modules']

    def test_specified_module_isnt_present_in_module_config_file(self):
        # remove_module removes mariadb, but tag config file only contains ant-1.10
        fake_tag_configs = {
            'f30-build': {
                'modules': [
                    {
                        'name': 'ant',
                        'stream': '1.10',
                        'requires': {'platform': 'f30'},
                        'buildrequires': {'platform': 'f30'},
                    }
                ],
                'owners': ['someone@example.com']
            }
        }
        self.write_tag_config_file(fake_tag_configs)

        self.remove_module(False)

        with open(self.tag_config_file, 'r') as f:
            tag_configs = json.load(f)

        expected = fake_tag_configs['f30-build']['modules']
        assert expected == tag_configs['f30-build']['modules']



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import json

from ursa_major.handlers.base import BaseHandler


class ShowConfigHandler(BaseHandler):
    def run(self):
        self.tag = self.args.tag
        self.tag_config_file = self.args.tag_config_file
        self.load_tag_config()

        if self.tag is None:
            content = json.dumps(self.tag_config, indent=4, sort_keys=True)
        else:
            config = self.tag_config.get(self.tag, None)
            if config is None:
                raise RuntimeError("Tag {} is not present in config file {}"
                                   .format(self.tag, self.tag_config_file))
            content = json.dumps(config, indent=4, sort_keys=True)
        print(content)



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

# Copied from MBS code
MBS_BUILD_STATES = {
    "init": 0,
    "wait": 1,
    "build": 2,
    "done": 3,
    "failed": 4,
    "ready": 5,
    "garbage": 6,
}


class ModuleConfigCmp(object):
    # User input module was not added to configuration file yet.
    NEW = 0
    # User input module has same config with the one in configuration file.
    # That means name is same, stream is same, and requires is also the
    # same.
    SAME = 1
    # User input module has different config from the one found in
    # configuration file.
    # Probably either stream or requires is different, or both
    DIFFERENT = 2


class ModuleConfig(object):
    """Class representing a module config"""

    def __init__(self, module_config):
        self._raw_module_config = module_config
        self.priority = module_config.get('priority')
        self.name = module_config['name']
        self.stream = module_config['stream']
        self.requires = module_config.get('requires', None)
        self.buildrequires = module_config.get('buildrequires', None)

    @property
    def raw_module_config(self):
        return self._raw_module_config

    def __str__(self):
        return '{}:{}:{!r}:{!r}'.format(
            self.name, self.stream, self.requires, self.buildrequires)

    def compare(self, other):
        """
        Compare user input module config with the one found from configuration
        file.
        """
        if other is None:
            return ModuleConfigCmp.NEW
        if self.priority != other.priority:
            return ModuleConfigCmp.DIFFERENT
        if self.requires != other.requires:
            return ModuleConfigCmp.DIFFERENT
        if self.buildrequires != other.buildrequires:
            return ModuleConfigCmp.DIFFERENT
        return ModuleConfigCmp.SAME



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

from ursa_major.handlers.base import BaseHandler
from ursa_major.logger import log


class RemoveModuleHandler(BaseHandler):
    def run(self):
        self.dry_run = self.args.dry_run

        self.tag = self.args.tag
        self.module_name = self.args.module_name
        self.module_stream = self.args.module_stream
        self.module_requires = {}
        if self.args.module_requires is not None:
            self.module_requires = dict(self.args.module_requires)
        if self.args.module_buildrequires is not None:
            self.module_buildrequires = dict(self.args.module_buildrequires)
        else:
            self.module_buildrequires = {}
        self.tag_config_file = self.args.tag_config_file
        self.not_update_config = self.args.not_update_config

        self.connect_koji()
        self.connect_mbs()

        self.load_tag_config()
        if self.tag not in self.tag_config.keys():
            log.warning("Tag %s is not present in tag config file %s",
                        self.tag, self.tag_config_file)

        try:
            tag_inheritance = self.koji.get_inheritance_data(self.tag)
        except Exception as e:
            raise RuntimeError("Unable to get inheritance data of tag {}:\n"
                               "{!s}".format(self.tag, e))

        inheritance_tags = [t['name'] for t in tag_inheritance]

        log.info("Querying modules of '%s:%s' from MBS with requires: %r and "
                 "buildrequires: %r",
                 self.module_name, self.module_stream, self.module_requires,
                 self.module_buildrequires)

        modules = self.mbs.get_modules(
            name=self.module_name,
            stream=self.module_stream,
            requires=self.module_requires,
            buildrequires=self.module_buildrequires)

        module_tags = [m['koji_tag'] for m in modules]
        match_tags = list(set(inheritance_tags) & set(module_tags))
        if match_tags:
            log.info("Found matched module tags under %s: %r", self.tag, match_tags)
            self.koji.remove_tags_from_inheritance(self.tag, match_tags)
        else:
            log.info("No match tag found in tag %s's inheritance data", self.tag)

        if self.not_update_config:
            log.info("Not update config is specified, not update config")
            return

        if self.tag not in self.tag_config.keys():
            log.info("Tag config file doesn't contain tag %s, not update config file",
                     self.tag)
            return

        module_configs = self.tag_config[self.tag].get('modules', [])

        found_config = None
        for config in module_configs:
            if all((
                self.module_name == config['name'],
                self.module_stream == config['stream'],

                # NOTE: we only find config that matches module_requires exactly
                self.module_requires == config.get('requires', {}),
                # Same as requires
                self.module_buildrequires == config.get('buildrequires', {})
            )):
                found_config = config
                break

        if found_config is None:
            log.info("Doesn't find match config under tag %s, not update config file",
                     self.tag)
        else:
            module_configs.remove(found_config)
            log.info("Updating tag config file to remove module:\n%r", found_config)
            self.write_tag_config()



# -*- coding: utf-8 -*-
# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

from ursa_major import MBS_BUILD_STATES
from ursa_major.handlers.base import BaseHandler
from ursa_major.logger import log


class UpdateTagHandler(BaseHandler):
    def __init__(self, *args, **kwargs):
        super(UpdateTagHandler, self).__init__(*args, **kwargs)
        self.dry_run = False
        self.wait_regen_repo = False
        self.inheritance_changed = False

    def run(self):
        self.debug = self.args.debug
        self.dry_run = self.args.dry_run
        self.wait_regen_repo = self.args.wait_regen_repo

        self.tag = self.args.tag
        self.tag_config_file = self.args.tag_config_file
        self.load_tag_config()

        if self.tag not in self.tag_config.keys():
            raise RuntimeError("Tag '%s' is not found in tag config file %s" % (self.tag,
                               self.tag_config_file))

        self.connect_koji(dry_run=self.dry_run)
        self.connect_mbs()

        self.tag_info = self.koji.get_tag(self.tag)
        if self.tag_info is None:
            raise RuntimeError("Tag '%s' is not found in koji (profile: %s)" % (self.tag,
                               self.koji.profile))

        modules = self.tag_config.get(self.tag).get('modules', {})
        if not modules:
            log.warning("No module specified for tag '%s' in tag config file", self.tag)
            return

        inheritance = self.koji.get_inheritance_data(self.tag)

        # convert inheritance list to dict for convenience
        # dict key is tag name and value is the inheritance data of that tag
        inheritance_dict = {}
        for tag in inheritance:
            inheritance_dict[tag['name']] = tag

        # for each module in config:
        # 1. check whether there is any ready build exists in MBS
        # 2. check whether the build tag found in step #1 is in inheritance data
        #    already and with same priority, if it is in inheritance but with
        #    different priority, update priority
        # 3. if build tag is not inheritance, add this tag into inheritance
        # 4. check whether there is any old build tags for this module (name+stream)
        #    exist in inheritance, if true, remove these tags from inheritance
        for module in modules:
            log.debug("Processing module in config: \n%s", str(module))
            module_name = module.get('name')
            module_stream = module.get('stream')
            module_priority = module.get('priority')
            module_requires = module.get('requires', None)
            module_buildrequires = module.get('buildrequires', None)

            latest_build_tag = None
            # find out the latest ready module for this config from MBS
            ready_modules = self.mbs.get_modules(
                buildrequires=module_buildrequires,
                requires=module_requires,
                name=module_name,
                stream=module_stream,
                state=MBS_BUILD_STATES['ready'],
                page=1)  # we only need the default first page items
            if ready_modules:
                latest_build_tag = ready_modules[0]['koji_tag']

            if latest_build_tag is None:
                log.warning("There is no ready build found for %s, skipping", str(module))
                continue

            has_latest_tag = latest_build_tag in inheritance_dict
            same_priority = False
            if has_latest_tag:
                same_priority = inheritance_dict[latest_build_tag]['priority'] == module_priority

            if has_latest_tag and same_priority:
                log.info("Tag '%s' is in inheritance data of %s with same priority, skipping",
                         latest_build_tag, self.tag)
                continue

            if has_latest_tag and not same_priority:
                log.info("Tag '%s' is in inhertiance data of %s but has different priority, "
                         "will update inheritance data with new priority",
                         latest_build_tag, self.tag)
                # update with new priority value
                inheritance_dict[latest_build_tag]['priority'] = module_priority
                self.inheritance_changed = True
                continue

            # after check, the latest build tag is not in inheritance, we'll add this
            # tag to inheritance
            latest_tag_info = self.koji.get_tag(latest_build_tag)
            if latest_tag_info is None:
                raise RuntimeError("Tag '%s' is not found in koji (profile: %s)" %
                                   (latest_build_tag, self.koji.profile))

            # this is inheritance data for the latest module build we're going to add
            module_tag_data = {}
            module_tag_data['child_id'] = self.tag_info['id']
            module_tag_data['parent_id'] = latest_tag_info.get('id')
            module_tag_data['name'] = latest_tag_info.get('name')
            module_tag_data['priority'] = module_priority
            module_tag_data['maxdepth'] = module.get('maxdepth', None)
            module_tag_data['intransitive'] = module.get('intransitive', False)
            module_tag_data['noconfig'] = module.get('noconfig', False)
            module_tag_data['pkg_filter'] = module.get('pkg_filter', '')

            # check any old module tag exist in inheritance, for each
            # name + stream combination, there is up to 1 tag can exists
            # in inheritance data. Remove all old tags.
            modules_with_name_stream = self.mbs.get_modules(
                name=module_name,
                stream=module_stream,
                state=[MBS_BUILD_STATES['ready'], MBS_BUILD_STATES['garbage']])

            old_build_tags = [m['koji_tag'] for m in modules_with_name_stream]

            inheritance_old_tags = set(old_build_tags) & set(inheritance_dict.keys())

            # add new inheritance data for this module build, update the inheritance
            # after checking the old tags in inheritance
            log.info("Adding tag '%s' to inheritance data", latest_build_tag)
            inheritance_dict[latest_build_tag] = module_tag_data
            self.inheritance_changed = True

            if len(inheritance_old_tags) == 0:
                # there is no old tag found for this module, do nothing
                continue

            if len(inheritance_old_tags) > 0:
                # found old tag(s) of this NAME:STREAM in inheritance, remove them
                for old_tag in inheritance_old_tags:
                    log.info("Removing tag '%s' from inheritance data", old_tag)
                    inheritance_dict[old_tag]['delete link'] = True

        # modules loop finished, now we have the finalized inheritance data
        if self.inheritance_changed:
            # here we convert the inheritance dict back to list
            inheritance_data = [v for k, v in inheritance_dict.items()]
            self.koji.login()
            log.info("Updating inheritance of tag %s with data: \n%s",
                     self.tag, str(inheritance_data))
            self.koji.set_inheritance_data(self.tag, inheritance_data)
            self.koji.regen_repo(self.tag, wait=self.wait_regen_repo)
        else:
            log.info("No change to inheritance data of tag '%s', skipping",
                     self.tag)



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Filip Valder <fvalder@redhat.com>
#            Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import logging
import smtplib

from email.message import Message
from email.header import Header
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.utils import formataddr, parseaddr
from ursa_major.utils import jinja2_env
from ursa_major.logger import log


class MailAPI:
    def __init__(self, config):
        """
        Instantiate Mail API (connect to mail server)
        """

        # Set coding
        self.coding = 'utf-8'

        # Load configuration
        self.conf = config

        # Configure Jinja2 environment
        self.env = jinja2_env('mail')

        # Configure sender's e-mail address
        f_realname, f_addr = parseaddr(self.conf['mail_from'])
        if not f_addr:
            raise Exception("Unable to parse sender's e-mail address. "
                            "Check 'mail_from' in configuration of ursa-major.")
        self.fromaddr = formataddr((f_realname, f_addr))

        # Set up mail log level
        log_level = self.conf.get('mail_log_level', 'NOTSET')
        self.log_level = getattr(logging, log_level.upper())

        # Set up mail processing
        self.processing = self._processing()

        # Initiate SMTP connection
        self.server = self._connect()

    def _processing(self):
        if 'mail_processing' not in self.conf:
            return False

        if self.conf['mail_processing'].lower() in ['enable', 'yes', 'true']:
            return True
        elif self.conf['mail_processing'].lower() in ['disable', 'no', 'false']:
            return False
        else:
            raise ValueError(
                "Unknown value for 'mail_processing' option: {}".format(
                    self.conf['mail_processing']))

    def _connect(self):
        """
        Connect to mail server
        """

        if not self.processing:
            return

        server = None

        address = self.conf['mail_server']
        try:
            server = smtplib.SMTP(address)
        except Exception:
            log.exception("Cannot connect to SMTP server at %r." % address)
            raise
        if not server:
            return

        try:
            server.starttls()
        except Exception:
            log.exception("STARTTLS extension not supported by server. "
                          "WARNING! Communication with the server will not be "
                          "secured.")

        server.ehlo_or_helo_if_needed()

        return server

    def _is_connected(self):
        """
        Helper for checking SMTP connection status
        """

        try:
            status = self.server.noop()[0]
        except Exception:  # smtplib.SMTPServerDisconnected
            status = -1
        return True if status == 250 else False

    def _mail_render(self, template, data=None):
        """
        Render mail body (plain text/HTML) using Jinja2 and provided data
        """

        template += ".j2"
        log.debug("Rendering template '%s'." % (template))
        text = self.env.get_template(template)
        if data:
            msg = text.render(data=data)
        else:
            msg = text.render()

        return msg

    def _process_recipients(self, *addrs):
        """
        Process recipients (mail addresses) and return properly formatted
        commar-separated list of addresses.
        """

        all_addrs = list()
        for addr in addrs:
            all_addrs += [addr] if isinstance(addr, str) else addr

        # process only non-empty addresses
        all_addrs = [a for a in all_addrs if a]

        for i, a in enumerate(all_addrs):
            realname, addr = parseaddr(a)
            if not addr:
                log.warning("Unable to parse recipient's e-mail address. "
                            "Please check: %s", a if a else '<missing-address>')
            all_addrs[i] = realname, addr

        return ', '.join([formataddr((r, a)) for r, a in all_addrs if a])

    def _mail_headers(self, *headers):
        """
        This method formats headers for message envelope. Typically 'From',
        'To' and 'Subject' which aren't included by default.

        :param headers: a 3-tuple of header name, its value and True/False
                        bool whether to apply encoding;
                        this arg may be repeated multiple times for different
                        headers
        :return: message headers as string
        """

        msg = Message()
        for header_name, header_value, use_encode in headers:
            if use_encode:
                msg[header_name] = Header(header_value, self.coding,
                                          header_name=header_name).encode()
            else:
                msg[header_name] = Header(header_value, header_name=header_name)
        return msg.as_string().rstrip('\n') + '\n'

    def send_mail(self, template, toaddrs, subject='', data=None, ccs='',
                  bccs='', replyto=None):
        """
        Send e-mail rendered using predefined Jinja2 template.

        :param template: Jinja2 mail template name omitting the .j2 suffix
        :param toaddrs: str or list identifying message recipient(s)
                        (To: header)
        :param subject: message subject; this will be prefixed with
                        a configured prefix (if non-empty)
        :param data: data to be processed/rendered in the template
        :param ccs: str or list identifying additional message recipient(s)
                    (Cc: header)
        :param bccs: str or list identifying additional message recipient(s)
                     (Bcc: header)
        :param replyto: str identifying address for replies (Reply-To: header)
        """

        # Check if we're still connected to the mail server
        # If not => reconnect!
        if self.processing and not self._is_connected():
            self.server = self._connect()

        # Process mail template
        content = self._mail_render(template, data)
        if template.endswith('.html'):
            msg = MIMEMultipart('alternative')
            msg.attach(MIMEText(content, 'html', self.coding))
        elif template.endswith('.txt'):
            msg = MIMEText(content, _charset=self.coding)
        else:
            log.error("Unrecognized template format.")
            return

        # Process recipient(s)
        toaddrs = self._process_recipients(toaddrs)
        mail_always_cc = self.conf.get('mail_always_cc', '')
        mail_always_cc = [e.strip() for e in mail_always_cc.split(',') if '@' in e]
        ccs = self._process_recipients(ccs, mail_always_cc)

        mail_always_bcc = self.conf.get('mail_always_bcc', '')
        mail_always_bcc = [e.strip() for e in mail_always_bcc.split(',') if '@' in e]
        bccs = self._process_recipients(bccs, mail_always_bcc)

        # Process subject
        if 'mail_subject_prefix' in self.conf and self.conf['mail_subject_prefix']:
            subject = self.conf['mail_subject_prefix'] + ' ' + subject

        # Prepare the most important headers (will be used
        # for S/MIME envelope later)
        mail_envelope = self._mail_headers(
            ('From', self.fromaddr, False), ('To', toaddrs, False),
            ('Subject', subject, True))

        if ccs:
            msg['Cc'] = Header(ccs, header_name='Cc')
        if bccs:
            msg['Bcc'] = Header(bccs, header_name='Bcc')
        msg['Reply-To'] = Header(
            replyto if replyto else self.conf.get('mail_replyto', self.fromaddr),
            header_name='Reply-To')

        mail = mail_envelope + msg.as_string()

        # To addresses should be a list, a bare string will be treated as a list
        # with 1 address, CCs and BCCs should be added in to to-addresses, the
        # distinction between TO, CC and BCC occurs only in the text headers.
        # At the SMTP level, everybody is a recipient.
        to_addrs = [n.strip() for n in
                    (toaddrs.split(',') + ccs.split(',') + bccs.split(',')) if n]

        try:
            if self.processing:
                self.server.sendmail(self.fromaddr, to_addrs, mail)
            log.log(self.log_level, mail)
        except smtplib.SMTPException:
            log.exception("Cannot send mail to recipients: %s" % (toaddrs))



# -*- coding: utf-8 -*-
# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import json
import os
import tempfile
import pytest

from mock import patch
from tests import TEST_DATA_DIR, MockMBSBuildsData, make_mmd, URSA_MAJOR_TEST_CONFIG
from ursa_major import ModuleConfig
from ursa_major.cli import main
from ursa_major.handlers.add_module import AddModuleHandler


def get_tag_side_effect(info):
    return {
        'f30-build': {'id': 99, 'name': 'f30-build'},
        'module-mariadb-10.4-3020190313091759-a5b0195c': {
            'id': 100,
            'name': 'module-mariadb-10.4-3020190313091759-a5b0195c',
        },
        'module-mariadb-10.4-3020190304180835-a5b0195c': {
            'id': 101,
            'name': 'module-mariadb-10.4-3020190304180835-a5b0195c',
        },
    }[info]


class TestAddModule:

    def setup_method(self, method):
        fd, self.tag_config_file = tempfile.mkstemp()
        os.close(fd)

        self.ClientSession_patcher = patch('koji.ClientSession')
        self.mock_ClientSession = self.ClientSession_patcher.start()
        self.koji_session = self.mock_ClientSession.return_value

        self.get_patcher = patch('requests.get')
        self.mock_get = self.get_patcher.start()

    def teardown_method(self, method):
        self.get_patcher.stop()
        self.ClientSession_patcher.stop()

        os.unlink(self.tag_config_file)

    def write_tag_config_file(self, content):
        with open(self.tag_config_file, 'w') as f:
            json.dump(content, f)

    def read_tag_config_file(self):
        with open(self.tag_config_file, 'r') as f:
            return json.load(f)

    def add_module(self,
                   command_options=None,
                   tag_configs_content=None,
                   update_inheritance_only=None,
                   update_tag_config_only=None,
                   mock_inheritance_data=None,
                   mock_mbs_modules=None):
        """
        Run add-module command for test. Behavior could be changed by arguments
        for specific tests.
        """
        self.koji_session.getInheritanceData.return_value = mock_inheritance_data or [
            {'parent_id': 20,
             'name': 'module-ant-1.10-20181122140939-819b5873',
             'priority': 10},
        ]
        self.koji_session.getTag.side_effect = get_tag_side_effect

        mock_builds = mock_mbs_modules or [
            {
                'koji_tag': 'module-mariadb-10.4-3020190313091759-a5b0195c',
                'modulemd': make_mmd(name='mariadb', stream='10.4',
                                     version='3020190313091759', context='a5b0195c',
                                     requires={'platform': 'f30'},
                                     buildrequires={'platform': 'f30'}).dumps(),
            },
            {
                'koji_tag': 'module-mariadb-10.4-3020190304180835-a5b0195c',
                'modulemd': make_mmd(name='mariadb', stream='10.4',
                                     version='3020190304180835', context='a5b0195c',
                                     requires={'platform': 'f30'},
                                     buildrequires={'platform': 'f30'}).dumps(),
            },
        ]
        mock_builds_data = MockMBSBuildsData(mock_builds)

        self.mock_get.side_effect = mock_builds_data.get

        # For this test, we write an empty tag config file.
        self.write_tag_config_file(tag_configs_content or {})

        cli_cmd_options = command_options or [
            '--name', 'mariadb', '--stream', '10.4', '--priority', '50',
            '--require', 'platform:f30',
            '--buildrequire', 'platform:f30',
            '--tag', 'f30-build',
            '--tag-config-file', self.tag_config_file
        ]
        if update_inheritance_only:
            cli_cmd_options.append('--update-inheritance-only')
        if update_tag_config_only:
            cli_cmd_options.append('--update-config-only')

        config_file = os.path.join(TEST_DATA_DIR, 'ursa-major-test.conf')
        cli_cmd = ['ursa-major', 'add-module', '--config', config_file] + cli_cmd_options

        with patch('sys.argv', new=cli_cmd):
            main()

    def test_common_add_module(self):
        """Test add a module to tag config file and its koji_tag to tag inheritance"""
        self.add_module()

        self.koji_session.setInheritanceData.assert_called_once_with(99, [
            {
                'parent_id': 20,
                'name': 'module-ant-1.10-20181122140939-819b5873',
                'priority': 10
            },
            {
                'child_id': 99,
                'parent_id': 100,
                'name': "module-mariadb-10.4-3020190313091759-a5b0195c",
                'priority': 50,
                'maxdepth': None,
                'intransitive': False,
                'noconfig': False,
                'pkg_filter': ''
            },
        ])

        expected = [{
            'name': 'mariadb',
            'stream': '10.4',
            'requires': {'platform': 'f30'},
            'buildrequires': {'platform': 'f30'},
            'priority': 50
        }]
        tag_configs = self.read_tag_config_file()
        assert expected == tag_configs['f30-build']['modules']

    def test_add_module_koji_tag_only(self):
        self.add_module(tag_configs_content={},
                        update_inheritance_only=True)

        self.koji_session.setInheritanceData.assert_called_once_with(99, [
            {
                'parent_id': 20,
                'name': 'module-ant-1.10-20181122140939-819b5873',
                'priority': 10
            },
            {
                'child_id': 99,
                'name': 'module-mariadb-10.4-3020190313091759-a5b0195c',
                'parent_id': 100,
                'priority': 50,
                'maxdepth': None,
                'intransitive': False,
                'noconfig': False,
                'pkg_filter': ''
            },
        ])

        tag_configs = self.read_tag_config_file()
        assert {} == tag_configs, 'Tag config file should keep unchanged.'

    def test_add_module_but_only_update_tag_config_file(self):
        self.add_module(tag_configs_content={},
                        update_tag_config_only=True)

        self.koji_session.setInheritanceData.assert_not_called()

        expected = [{
            'name': 'mariadb',
            'stream': '10.4',
            'requires': {'platform': 'f30'},
            'buildrequires': {'platform': 'f30'},
            'priority': 50
        }]

        tag_configs = self.read_tag_config_file()
        assert expected == tag_configs['f30-build']['modules']

    @patch('ursa_major.handlers.add_module.AddModuleHandler.write_tag_config')
    def test_specified_same_module_config(
            self, write_tag_config):
        """
        Both tag inheritance and tag config file are not updated because same
        module metadata is specified from command line.
        """
        self.add_module(tag_configs_content={
            'f30-build': {
                'modules': [
                    # Compare with the metadata specified from command line,
                    # all of them are same.
                    {
                        'name': 'mariadb',
                        'stream': '10.4',
                        'requires': {'platform': 'f30'},
                        'buildrequires': {'platform': 'f30'},
                        'priority': 50,
                    }
                ],
                'owners': ['owner@example.com']
            }
        })

        self.koji_session.setInheritanceData.assert_not_called()
        write_tag_config.assert_not_called()

    def test_remove_old_koji_tag_and_add_latest_koji_tag_to_inheritance(self):
        self.add_module(mock_inheritance_data=[
            # Noisy koji_tag, which should not be changed.
            {
                'parent_id': 5,
                'name': 'module-ant-1.10-20181122140939-819b5873',
                'priority': 2,
            },
            # This tag will be removed because this old koji_tag exists in the modules
            # returned from MBS.
            {
                'parent_id': 101,
                'name': 'module-mariadb-10.4-3020190304180835-a5b0195c',
                'priority': 10,
            },
        ])

        self.koji_session.setInheritanceData.assert_called_once_with(99, [
            {
                'parent_id': 5,
                'name': 'module-ant-1.10-20181122140939-819b5873',
                'priority': 2,
            },
            {
                'parent_id': 101,
                'name': 'module-mariadb-10.4-3020190304180835-a5b0195c',
                'priority': 10,
                'delete link': True,
            },
            {
                'child_id': 99,
                'parent_id': 100,
                'name': 'module-mariadb-10.4-3020190313091759-a5b0195c',
                'priority': 50,
                'maxdepth': None,
                'intransitive': False,
                'noconfig': False,
                'pkg_filter': ''
            },
        ])

    def test_add_module_with_buildrequires(self):
        """
        Add a module with buildrequires specified from command line but no
        buildrequires was specified before and the corresponding module config
        does not have buildrequires in tag config file.

        The expected result is to remove the koij_tag of the exisitng module
        within tag config file, and find latest module with the specified
        buildrequires by CLI option --buildrequires, and add its koji_tag to
        tag inheritance eventually.
        """
        self.add_module(
            mock_inheritance_data=[
                # Noisy tag in inheritance which should not be changed.
                {
                    'parent_id': 20,
                    'name': 'module-ant-1.10-20181122140939-819b5873',
                    'priority': 10,
                },
                # The is is the koji_tag that was added when buildrequires is
                # not specified in tag config. So, it will be removed and
                # koji_tag of latest mariadb:10.4 will be added to tag
                # inheritance.
                {
                    'parent_id': 101,
                    'name': 'module-mariadb-10.4-3020190304180835-a5b0195c',
                    'priority': 20,
                },
            ],
            tag_configs_content={
                'f30-build': {
                    'modules': [
                        # Compare with the metadata specified from command line,
                        # all of them are same.
                        {
                            'name': 'mariadb',
                            'stream': '10.4',
                            'requires': {'platform': 'f30'},
                            'priority': 50,
                            # Note that, this module was added without
                            # specifying buildrequires, whereas it is given by
                            # option --buildrequire from command line. This is
                            # just the difference.
                        }
                    ],
                    'owners': ['owner@example.com']
                }
            },
        )

        self.koji_session.setInheritanceData.assert_called_once_with(99, [
            {
                'parent_id': 20,
                'name': 'module-ant-1.10-20181122140939-819b5873',
                'priority': 10,
            },
            {
                'parent_id': 101,
                'name': 'module-mariadb-10.4-3020190304180835-a5b0195c',
                'priority': 20,
                'delete link': True,
            },
            # This is the koji_tag of the latest module mariadb:10.4 to add
            {
                'child_id': 99,
                'parent_id': 100,
                'name': 'module-mariadb-10.4-3020190313091759-a5b0195c',
                'priority': 50,
                'maxdepth': None,
                'intransitive': False,
                'noconfig': False,
                'pkg_filter': ''
            },
        ])

        expected = {
            'f30-build': {
                'modules': [
                    {
                        'name': 'mariadb',
                        'stream': '10.4',
                        'requires': {'platform': 'f30'},
                        'buildrequires': {'platform': 'f30'},
                        'priority': 50,
                    }
                ],
                'owners': ['owner@example.com']
            }
        }
        assert expected == self.read_tag_config_file()


class TestAddModuleFunctions:

    # NOTE: once MBS supports filtering modules builds by requires or
    # buildrequires, the test data should be updated because it is unnecessary
    # to filter module builds locally by Ursa-Major itself.

    @pytest.mark.parametrize(
        'module_config,module_builds,expected_koji_tags', [
            # None of these koji_tags exist in tag inheritance.
            (ModuleConfig({'name': 'mariadb', 'stream': '10.3',
                           'version': '3120190304180825', 'context': 'f636be4b'}),
             [{'id': 1,
               'name': 'mariadb',
               'koji_tag': 'module-mariadb-10.3-3020190301191548-a5b0195c'},
              {'id': 2,
               'name': 'mariadb',
               'koji_tag': 'module-mariadb-10.3-3120190304180825-f636be4b'}],
             []),

            # koji_tag exists in tag inheritance
            (ModuleConfig({'name': 'mariadb', 'stream': '10.4',
                           'version': '3120190304180825', 'context': 'f636be4b'}),
             [{'id': 1,
               'name': 'mariadb',
               'koji_tag': 'module-mariadb-10.4-3020190313091759-a5b0195c'},
              {'id': 2,
               'name': 'mariadb',
               # This older koji_tag was added previously
               'koji_tag': 'module-mariadb-10.4-3020190304180835-a5b0195c'}],
             ['module-mariadb-10.4-3020190304180835-a5b0195c']),

            # Modules are filtered by requires.
            # None of these koji_tags exist in tag inheritance.
            (ModuleConfig({'name': 'mariadb', 'stream': '10.3',
                           'version': '3120190304180825', 'context': 'f636be4b',
                           'requires': {'platform': 'f30'}}),
             [{'id': 1,
               'name': 'mariadb',
               'koji_tag': 'module-mariadb-10.3-3020190301191548-a5b0195c',
               'modulemd': make_mmd(name='mariadb', stream='10.3',
                                    version='3020190301191548', context='a5b0195c',
                                    requires={'platform': 'f30'},
                                    buildrequires={'platform': 'f30'}).dumps(),
               },
              {'id': 2,
               'name': 'mariadb',
               'koji_tag': 'module-mariadb-10.3-3120190304180825-f636be4b',
               'modulemd': make_mmd(name='mariadb', stream='10.3',
                                    version='3120190304180825', context='f636be4b',
                                    requires={'platform': 'f29'},
                                    buildrequires={'platform': 'f29'}).dumps(),
               }],
             []),

            # koji_tag exists in tag inheritance and module builds
            (ModuleConfig({'name': 'mariadb', 'stream': '10.4',
                           'version': '3120190304180825', 'context': 'f636be4b',
                           'requires': {'platform': 'f30'},
                           'buildrequires': {'platform': 'f30'}}),
             [{'id': 1,
               'name': 'mariadb',
               'koji_tag': 'module-mariadb-10.4-3020190313091759-a5b0195c',
               'modulemd': make_mmd(name='mariadb', stream='10.4',
                                    version='3020190313091759', context='a5b0195c',
                                    requires={'platform': 'f29'},
                                    buildrequires={'platform': 'f29'}).dumps(),
               },
              {'id': 2,
               'name': 'mariadb',
               # This is the koji_tag for platform:f30
               'koji_tag': 'module-mariadb-10.4-3020190304180835-a5b0195c',
               'modulemd': make_mmd(name='mariadb', stream='10.4',
                                    version='3020190304180835', context='a5b0195c',
                                    requires={'platform': 'f30'},
                                    buildrequires={'platform': 'f30'}).dumps(),
               }],
             ['module-mariadb-10.4-3020190304180835-a5b0195c']),

            # filter with requires/buildrequires, it won't filter out mariadb:10.4 modules
            (ModuleConfig({'name': 'mariadb', 'stream': '10.4',
                           'version': '3120190304180825', 'context': 'f636be4b',
                           'requires': {'platform': 'f30'},
                           'buildrequires': {'platform': 'f30'}}),
             [{'id': 1,
               'name': 'mariadb',
               'koji_tag': 'module-mariadb-10.4-3020190313091759-a5b0195c',
               'modulemd': make_mmd(name='mariadb', stream='10.4',
                                    version='3020190313091759', context='a5b0195c',
                                    requires={'platform': 'f29'},
                                    buildrequires={'platform': 'f29'}).dumps(),
               },
              # Although this koji_tag was added to tag inheritance, but no
              # koji_tag is found that matches platform:f30.
              {'id': 2,
               'name': 'mariadb',
               'koji_tag': 'module-mariadb-10.4-3020190304180835-a5b0195c',
               'modulemd': make_mmd(name='mariadb', stream='10.4',
                                    version='3020190304180835', context='a5b0195c',
                                    requires={'platform': 'f28'},
                                    buildrequires={'platform': 'f28'}).dumps(),
               }],
             ['module-mariadb-10.4-3020190304180835-a5b0195c']),
        ])
    @patch('ursa_major.koji_service.koji')
    @patch('requests.get')
    def test_find_module_tags_from_koji_inheritance(
            self, get, koji, module_config, module_builds, expected_koji_tags):
        koji_session = koji.ClientSession.return_value
        koji_session.getInheritanceData.return_value = [
            {'parent_id': 100,
             'name': 'module-ant-1.10-20181122140939-819b5873'},
            {'parent_id': 101,
             'name': 'module-hyperfine-latest-3120190318171218-f636be4b'},
            {'parent_id': 102,
             'name': 'module-exa-latest-2820190316115637-8de94f32'},
            {'parent_id': 103,
             'name': 'module-mariadb-10.4-3020190304180835-a5b0195c'},
        ]

        mock_mbs_builds_data = MockMBSBuildsData(module_builds)
        get.side_effect = mock_mbs_builds_data.get

        handler = AddModuleHandler(config=URSA_MAJOR_TEST_CONFIG)
        handler.connect_koji()
        handler.connect_mbs()
        found_koji_tags = handler.find_module_tags_from_koji_inheritance(
            'f30-build', module_config)
        assert expected_koji_tags == found_koji_tags



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import argparse
import json
import logging
import os
import re
import sys
import six

from ursa_major.logger import log
from ursa_major.handlers.add_module import AddModuleHandler
from ursa_major.handlers.add_tag import AddTagHandler
from ursa_major.handlers.check_config import CheckConfigHandler
from ursa_major.handlers.remove_module import RemoveModuleHandler
from ursa_major.handlers.show_config import ShowConfigHandler
from ursa_major.handlers.update_tag import UpdateTagHandler

if six.PY3:  # SafeConfigParser == ConfigParser, former deprecated in >= 3.2
    from six.moves.configparser import ConfigParser
else:
    from six.moves.configparser import SafeConfigParser as ConfigParser

cli_name = os.path.basename(sys.argv[0])
cwd = os.getcwd()

DEFAULT_CONFIG_FILE = "/etc/ursa-major/%s.conf" % cli_name

DEFAULT_USER_CONFIG_FILE = os.path.join(
    os.path.expanduser('~'), '.config', 'ursa-major', '%s.conf' % cli_name)

DEFAULT_TAG_CONFIG_FILE = os.path.join(cwd, 'ursa-major.json')


def validate_json_file(value):
    if not os.path.exists(value):
        if value == DEFAULT_TAG_CONFIG_FILE:
            raise argparse.ArgumentTypeError(
                'Default tag config file ursa-major.json is not found from '
                'current directory.\n'
                'Make sure to run ursa-major alongside it, otherwise specify '
                'the file by option --tag-config-file.')
        else:
            raise argparse.ArgumentTypeError(
                'Specified tag config file does not exist: {}'.format(value))

    try:
        with open(value, 'r') as f:
            json.load(f)
    except IOError as e:
        raise argparse.ArgumentTypeError('Cannot read tag config file: {}'.format(str(e)))
    except Exception as e:
        raise argparse.ArgumentTypeError("Invalid json file {}: {}.".format(value, str(e)))
    return value


def clean_module_require(value):
    ore = re.compile(r'^(\w+):([-.\w]+)$')
    match = ore.match(value)
    if match:
        return match.groups()
    raise argparse.ArgumentTypeError('{} is not a pair of NAME:STREAM.'.format(value))


def main(args=None):
    parser = argparse.ArgumentParser()
    subparser = parser.add_subparsers(help="Sub commands", dest='subcommand')
    subparser.required = True

    global_args_parser = argparse.ArgumentParser(add_help=False)
    global_args_parser.add_argument(
        '--dry-run',
        action='store_true',
        default=False,
        help='dry run mode')
    global_args_parser.add_argument(
        '--debug',
        action='store_true',
        default=False,
        help='show debug log')
    global_args_parser.add_argument(
        '--config',
        metavar='PATH',
        default=DEFAULT_CONFIG_FILE,
        help='config file for ursa-major')
    global_args_parser.add_argument(
        '--user-config',
        metavar='PATH',
        default=DEFAULT_USER_CONFIG_FILE,
        help='user config file to use, can override default configurations')
    global_args_parser.add_argument(
        '--tag-config-file',
        metavar='PATH',
        default=DEFAULT_TAG_CONFIG_FILE,
        type=validate_json_file,
        help='tag config file, default to "ursa-major.json" in current working directory')
    global_args_parser.add_argument(
        '--traceback',
        action='store_true',
        default=False,
        help='Print traceback if something wrong that ursa-major cannot do the job.')

    module_config_parser = argparse.ArgumentParser(add_help=False)
    module_config_parser.add_argument(
        '--name',
        metavar='NAME',
        dest='module_name',
        required=True,
        help="module name (required)")
    module_config_parser.add_argument(
        '--stream',
        metavar='STREAM',
        dest='module_stream',
        required=True,
        help="module stream (required)")
    module_config_parser.add_argument(
        '--require',
        metavar='NAME:STREAM',
        type=clean_module_require,
        action='append',
        dest='module_requires',
        help='module runtime dependency in NAME:STREAM')
    module_config_parser.add_argument(
        '--buildrequire',
        metavar='NAME:STREAM',
        type=clean_module_require,
        action='append',
        dest='module_buildrequires',
        help='module build dependency in NAME:STREAM, for example platform:f30')

    show_config_parser = subparser.add_parser(
        "show-config",
        parents=[global_args_parser],
        help="show tag configuration content")
    show_config_parser.set_defaults(_class=ShowConfigHandler, func='run')
    show_config_parser.add_argument(
        '--tag',
        metavar='TAG',
        help="show config of this tag only")

    check_config_parser = subparser.add_parser(
        "check-config",
        parents=[global_args_parser],
        help="validate tag config file")
    check_config_parser.set_defaults(_class=CheckConfigHandler, func='run')
    check_config_parser.add_argument(
        '--allow-non-ready-module',
        action='store_true',
        default=False,
        help='allow non ready modules in config file')

    remove_module_parser = subparser.add_parser(
        "remove-module",
        parents=[global_args_parser, module_config_parser],
        help="remove a module config from a tag and also remove that module's "
             "koji_tag if it inherits from that tag.")
    remove_module_parser.set_defaults(_class=RemoveModuleHandler, func='run')
    remove_module_parser.add_argument(
        '--tag',
        metavar='TAG',
        required=True,
        help="a tag name (required). Specified module config will be removed "
             "if tag presents in tag config file and has the module config. "
             "The module's koji_tag will also be removed from tag inheritance "
             "if it inherits from this tag.")
    remove_module_parser.add_argument(
        '--not-update-config',
        action='store_true', default=False,
        help='do not update tag configuration file')

    add_module_parser = subparser.add_parser(
        'add-module',
        parents=[global_args_parser, module_config_parser],
        help="add a module to tag config file, find out the latest built "
             "module in MBS and add its tag to koji tag inheritance.")
    add_module_parser.set_defaults(_class=AddModuleHandler, func='run')
    add_module_parser.add_argument(
        '--tag',
        metavar='TAG',
        required=True,
        help="koji tag we operate against (required)")
    add_module_parser.add_argument(
        '--force',
        action='store_true',
        default=False,
        help="force to remove existing old koji tags from specified koji tag's "
             "inheritance data even there is no module build found in MBS for "
             "the specified module, and update config file as well")
    add_module_parser.add_argument(
        '--priority',
        type=int,
        required=True,
        help="priority of the module's koji tag in tag inheritance (required)")
    add_module_parser.add_argument(
        '--wait-regen-repo',
        action='store_true', default=False,
        help='wait for regen-repo task(s) to finish')
    mutual_args = add_module_parser.add_mutually_exclusive_group()
    mutual_args.add_argument(
        '--update-inheritance-only',
        action='store_true',
        default=False,
        help="only update the specified koji tag's inheritance data to add or "
             "edit the latest module build's tag, do not update tag configuration file")
    mutual_args.add_argument(
        '--update-config-only',
        action='store_true',
        default=False,
        help="only add the specified module into tag configuration file, do not "
             "update the specified koji tag's inheritance data")

    add_tag_parser = subparser.add_parser(
        "add-tag",
        parents=[global_args_parser],
        help="read module stage change message from an environment variable and "
             "add the module's tag to koji tag inheritance depending on the "
             "tag configuration")
    add_tag_parser.set_defaults(_class=AddTagHandler, func='run')
    add_tag_parser.add_argument(
        '--module-from-env',
        metavar='ENV_VAR',
        default='CI_MESSAGE',
        help="the environment variable contains the module state change message, "
             "default variable is 'CI_MESSAGE'")
    add_tag_parser.add_argument(
        '--send-mail',
        action='store_true', default=False,
        help='send mail after operations')
    add_tag_parser.add_argument(
        '--wait-regen-repo',
        action='store_true', default=False,
        help='wait for regen-repo task to finish')

    update_tag_parser = subparser.add_parser(
        'update-tag',
        parents=[global_args_parser],
        help=" "
             "module in MBS and add its tag to koji tag inheritance.")
    update_tag_parser.set_defaults(_class=UpdateTagHandler, func='run')
    update_tag_parser.add_argument(
        '--tag',
        metavar='TAG',
        required=True,
        help="koji tag to update inheritance data (required)")
    update_tag_parser.add_argument(
        '--wait-regen-repo',
        action='store_true', default=False,
        help='wait for regen-repo task(s) to finish')

    args = parser.parse_args(args)
    debug = getattr(args, 'debug', False)
    if debug:
        [h.setLevel(logging.DEBUG) for h in log.handlers]
        log.setLevel(logging.DEBUG)

    config = ConfigParser()
    sys_config = getattr(args, 'config', None)
    user_config = getattr(args, 'user_config', None)

    # read system config file
    if sys_config is not None:
        if os.path.exists(sys_config):
            log.info("Reading config from %s", sys_config)
            config.read(sys_config)
        else:
            log.error("Default ursa-major config file '%s' is missing", sys_config)
            sys.exit(1)

    # read user config file, which can overrides configurations
    # in system config file
    if user_config is not None and os.path.exists(user_config):
        log.info("Reading user config from %s", user_config)
        config.read(user_config)

    _class = args._class(config)
    _class.set_args(args)
    func = getattr(_class, args.func)
    try:
        func()
    except KeyboardInterrupt:
        sys.exit(0)
    except Exception as e:
        if args.traceback:
            log.exception(str(e))
        else:
            log.error(str(e))
        sys.exit(1)



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import requests

from ursa_major.utils import load_mmd, mmd_has_requires, mmd_has_buildrequires


class MBS(object):
    def __init__(self, server_url, api_version=1):
        self.server_url = server_url

        self.module_builds_api_url = (
            self.server_url.rstrip('/') +
            '/module-build-service/{!s}/module-builds/'.format(api_version))

    def get_module_mmd(self, build_id):
        """
        Create and return Modulemd.Module object for a module build.

        :param build_id: module build id
        :return: Modulemd.Module object
        """
        module_url = self.module_builds_api_url + str(build_id)
        params = {'verbose': 'true'}
        resp = requests.get(module_url, params=params)
        if resp.status_code != 200:
            resp.raise_for_status()

        module = resp.json()
        mmd = load_mmd(module['modulemd'])
        return mmd

    def module_has_requires(self, build_id, requires):
        """
        Check whether a module has requires.

        :param build_id: module build id
        :param requires: dict of requires, example:
            {'platform': 'f28', 'python3': 'master'}
        :return: True or False
        """
        mmd = self.get_module_mmd(build_id)
        return mmd_has_requires(mmd, requires)

    def get_modules(self, **kwargs):
        """
        Query MBS modules with specified query parameters.

        :param dict kwargs: query parameters in keyword arguments for listing
            module builds. Modules can be filtered by either requires or
            buildrequires or both by passing ``requires`` and ``buildrequires``,
            which accepts a mapping containing requirements, for example,
            ``{"platform": "f30"}``.
        :return: a list of mappings each of them presents a module returned
            from MBS without modified.
        :rtype: list[dict]
        """
        # MBS does not support filtering module builds based on requires and
        # buildrequires. Pop either of them and filter modules later by
        # ourselves.
        requires = kwargs.pop('requires', None)
        buildrequires = kwargs.pop('buildrequires', None)
        page = kwargs.get('page', None)
        per_page = kwargs.get('per_page', None)

        inspect_modulemd = False

        # get module builds from all pages by default
        all_pages = True
        # if page or per_page in query params, just get modules as requested
        if not (page is None and per_page is None):
            all_pages = False

        # we need modulemd if requires or buildrequires is in query params
        if requires or buildrequires:
            inspect_modulemd = True
            kwargs.update({'verbose': 'true'})

        resp = requests.get(self.module_builds_api_url, params=kwargs)
        if resp.status_code != 200:
            resp.raise_for_status()

        resp_data = resp.json()
        modules = resp_data['items']
        result_pages = resp_data['meta']['pages']
        result_total = resp_data['meta']['total']

        # the module builds are in multiple pages and we want all
        # modules, query with per_page = result_total
        if all_pages and result_pages > 1:
            kwargs['per_page'] = result_total
            resp = requests.get(self.module_builds_api_url, params=kwargs)
            if resp.status_code != 200:
                resp.raise_for_status()

            resp_data = resp.json()
            modules = resp_data['items']

        if inspect_modulemd:
            matched = []
            for module in modules:
                mmd = load_mmd(module['modulemd'])
                b = True
                if requires:
                    b = b and mmd_has_requires(mmd, requires)
                if buildrequires:
                    b = b and mmd_has_buildrequires(mmd, buildrequires)
                if b:
                    matched.append(module)
            return matched
        return modules



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import mock

from tests import UrsaMajorTestCase, MockResponse, MockMBSBuildsData
from ursa_major.mbs import MBS


class TestMBS(UrsaMajorTestCase):
    @mock.patch('ursa_major.mbs.requests.get')
    def test_get_modules_with_no_page_releated_params(self, get):
        mock_builds = [
            {'id': 1, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '123', 'state': 5},
            {'id': 2, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '234', 'state': 5},
            {'id': 3, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '456', 'state': 5},
            {'id': 4, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '789', 'state': 5},
            {'id': 5, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '012', 'state': 5},
        ]

        mock_data = MockMBSBuildsData(mock_builds)

        get.side_effect = mock_data.get

        mbs = MBS('http://mbs.example.com')
        modules = mbs.get_modules(name='mariadb')

        self.assertEqual(modules, mock_builds)

    @mock.patch('ursa_major.mbs.requests.get')
    def test_get_modules_with_page_related_params(self, get):
        mock_builds = [
            {'id': 1, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '123', 'state': 5},
            {'id': 2, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '234', 'state': 5},
            {'id': 3, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '456', 'state': 5},
            {'id': 4, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '789', 'state': 5},
            {'id': 5, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '012', 'state': 5},
        ]

        mock_data = MockMBSBuildsData(mock_builds)

        get.side_effect = mock_data.get

        mbs = MBS('http://mbs.example.com')
        modules = mbs.get_modules(name='mariadb', page=1, per_page=2)

        expected = [
            {'id': 1, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '123', 'state': 5},
            {'id': 2, 'name': 'mariadb', 'stream': '10.4', 'koji_tag': '234', 'state': 5},
        ]

        self.assertEqual(modules, expected)

    @mock.patch('ursa_major.mbs.requests.get')
    def test_get_module_mmd(self, get):

        testmodule_file = 'mbs_testmodule_604_verbose.json'
        resp_data = self.load_json_from_file(testmodule_file)
        get.return_value = MockResponse(resp_data, 200)

        mbs = MBS('http://mbs.example.com')
        mmd = mbs.get_module_mmd(604)
        self.assertEqual(mmd.peek_name(), 'testmodule')
        self.assertEqual(mmd.peek_stream(), 'rhel-8.0')
        self.assertEqual(str(mmd.peek_version()), '20180507102412')
        self.assertEqual(mmd.peek_context(), 'c2c572ec')

    @mock.patch('ursa_major.mbs.requests.get')
    def test_module_has_requires(self, get):

        testmodule_file = 'mbs_testmodule_604_verbose.json'
        resp_data = self.load_json_from_file(testmodule_file)
        get.return_value = MockResponse(resp_data, 200)

        mbs = MBS('http://mbs.example.com')
        self.assertTrue(mbs.module_has_requires(604, {'platform': 'el8'}))

    @mock.patch('ursa_major.mbs.requests.get')
    def test_module_not_has_requires(self, get):

        testmodule_file = 'mbs_testmodule_604_verbose.json'
        resp_data = self.load_json_from_file(testmodule_file)
        get.return_value = MockResponse(resp_data, 200)

        mbs = MBS('http://mbs.example.com')
        self.assertFalse(mbs.module_has_requires(604, {'platform': 'f28'}))

    def test_get_modules(self):
        mbs_resp_data = self.load_json_from_file('test_get_modules_with_requires.json')
        mbs = MBS('http://mbs.example.com')

        # name, stream, requires, buildrequires, expected module build ids
        test_matrix = (
            ('testmodule', 'rhel-8.0', None, None, [604, 605]),
            ('testmodule', 'rhel-8.0', {'platform': 'el8'}, None, [604, 605]),
            ('testmodule', 'rhel-8.0', None, {'platform': 'el8'}, [604]),
            ('testmodule', 'rhel-8.0', {'platform': 'el8'}, {'platform': 'el8'}, [604]),
            ('testmodule', 'rhel-8.0', {'platform': 'f30'}, None, []),
            ('testmodule', 'rhel-8.0', {'platform': 'f30'}, {'platform': 'f30'}, []),
            ('testmodule', 'rhel-8.0', None, {'platform': 'f30'}, []),
        )

        for name, stream, requires, buildrequires, expected_build_ids in test_matrix:
            with mock.patch('ursa_major.mbs.requests.get') as get:
                get.return_value = MockResponse(mbs_resp_data, 200)

                module_builds = mbs.get_modules(
                    name=name,
                    stream=stream,
                    requires=requires,
                    buildrequires=buildrequires)

                params = {'name': name, 'stream': stream}
                if requires or buildrequires:
                    params['verbose'] = 'true'
                get.assert_called_once_with(mbs.module_builds_api_url, params=params)

                build_ids = sorted(build_info['id'] for build_info in module_builds)
                assert expected_build_ids == build_ids



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

""" Various utilities """
import os
import re
import gi
gi.require_version('Modulemd', '1.0')  # noqa
from gi.repository import Modulemd

from jinja2 import Environment, PackageLoader
from ursa_major.logger import log


def validate_mail_address(addr):
    """
    Validate e-mail address format.
    """

    mail_pattern = r"^[_a-z0-9-]+(.[_a-z0-9-]+)*@[a-z0-9-]+(.[a-z0-9-]+)*(.[a-z]{2,4})$"
    match = re.match(mail_pattern, addr)
    if match:
        return True

    log.error("E-mail address <%s> is not valid.", addr)
    return False


def jinja2_env(category):
    env = Environment(
        loader=PackageLoader("ursa_major", "templates/" + category)
    )

    def regex_replace(s, find, replace):
        return re.sub(find, replace, s)
    env.filters['regex_replace'] = regex_replace

    return env


def get_env_var(var, raise_if_not_exist=False):
    """
    Get variable value from os environ

    :param raise_if_not_exist: If true, raise ValueError when variable is
        not present in os environ, otherwise return None.
    """
    if var not in os.environ and raise_if_not_exist:
        raise ValueError("ENV variable '{}' does not exist.".format(var))
    return os.environ.get(var, None)


def load_mmd(yaml, is_file=False):
    """ Create Modulemd.Module object from string or file and return. """
    try:
        if is_file:
            mmd = Modulemd.Module().new_from_file(yaml)
        else:
            mmd = Modulemd.Module().new_from_string(yaml)
        # If the modulemd was v1, it will be upgraded to v2
        mmd.upgrade()
    except Exception:
        error = 'Invalid modulemd: {0}'.format(yaml)
        log.error(error)
        raise

    return mmd


def requires_included(mmd_requires, config_requires):
    """Test if requires defined in config is included in module metadata

    :param mmd_requires: a mapping representing either buildrequires or
        requires, which is returned from function ``peek_requires`` and
        ``peek_buildrequires``.
    :type mmd_requires: dict[str, Modulemd.SimpleSet]
    :param dict config_requires: a mapping representing either buildrequires or
        requires defined in config file. This is what to check if it is
        included in ``mmd_requires``.
    :return: True if all requires inside ``config_requires`` are included in
        module metadata. Otherwise, False is returned.
    :rtype: bool
    """
    for req_name, req_streams in config_requires.items():
        if req_name not in mmd_requires:
            return False

        if not isinstance(req_streams, list):
            req_streams = [req_streams]

        neg_reqs = set(s[1:] for s in req_streams if s.startswith('-'))
        pos_reqs = set(s for s in req_streams if not s.startswith('-'))

        streams = set(mmd_requires[req_name].dup())
        if streams & neg_reqs:
            return False
        if pos_reqs and not (streams & pos_reqs):
            return False
    return True


def mmd_has_requires(mmd, requires):
    """
    Check whether a module represent by the mmd has requires.

    :param mmd: Modulemd.Module object
    :param requires: dict of requires, example:
        {'platform': 'f28', 'python3': 'master'}
    """
    deps_list = mmd.peek_dependencies()
    mmd_requires = deps_list[0].peek_requires() if deps_list else {}
    return requires_included(mmd_requires, requires)


def mmd_has_buildrequires(mmd, config_buildrequires):
    """
    Check if a module metadata represented by the mmd has buildrequires.

    :param mmd: a module metadata.
    :type mmd: Modulemd.Module
    :param dict config_buildrequires: a mapping of buildrequires defined in
        config file to match the module metadata, for example:
        ``{'platform': 'f28', 'python3': 'master'}``.
    :return: True if the specified module metadata has the buildrequires
        defined in config.
    :rtype: bool
    """
    deps_list = mmd.peek_dependencies()
    mmd_requires = deps_list[0].peek_buildrequires() if deps_list else {}
    return requires_included(mmd_requires, config_buildrequires)



# -*- coding: utf-8 -*-

import os
from setuptools import find_packages, setup


def read_requirements(filename):
    specifiers = []
    dep_links = []

    with open(filename, 'r') as f:
        for line in f:
            if line.startswith('-r') or line.strip() == '':
                continue
            if line.startswith('git+'):
                dep_links.append(line.strip())
            else:
                specifiers.append(line.strip())

    return specifiers, dep_links


def get_long_description():
    with open('README.rst', 'r') as f:
        long_description = f.read()
    with open('CHANGELOG.rst', 'r') as f:
        long_description += '\n\n' + f.read()
    return long_description

setup_py_path = os.path.dirname(os.path.realpath(__file__))
requirements_file = os.path.join(setup_py_path, 'requirements.txt')
test_requirements_file = os.path.join(setup_py_path, 'test-requirements.txt')
install_requires, deps_links = read_requirements(requirements_file)
tests_require, _ = read_requirements(test_requirements_file)
if _:
    deps_links.extend(_)


setup(
    name='ursa-major',
    description="A utility for managing module tags in Koji's tag inheritance",
    long_description=get_long_description(),
    version='0.3.3',
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities',
    ],
    keywords='ursa-major modularity koji fedora',
    author='The Factory 2.0 Team',
    author_email='ursa-major-owner@fedoraproject.org',
    url='https://pagure.io/ursa-major/',
    license='MIT',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    dependency_links=deps_links,
    entry_points={
        'console_scripts': [
            'ursa-major = ursa_major.cli:main',
            'ursa-major-stage = ursa_major.cli:main'
        ],
    },
    data_files=[
        ('/etc/ursa-major/', ['conf/ursa-major.conf',
                              'conf/ursa-major-stage.conf']),
    ],
)
